from unittest import mock

import astropy.units as u
import numpy as np
import pytest
from astropy.time import Time
from astropy.tests.helper import assert_quantity_allclose

from sunpy.coordinates.frames import Helioprojective, HeliographicStonyhurst

from stixpy.frames import STIXImagingFrame


@pytest.mark.remote_data
def test_transforms_with_know_values():
    obstime = Time('2021-05-22T02:52:00')
    earth_coord = Helioprojective(-395 * u.arcsec, 330 * u.arcsec, observer='earth', obstime=obstime)
    earth_to_stix = earth_coord.transform_to(STIXImagingFrame(obstime=obstime))

    stix_coord = STIXImagingFrame(earth_to_stix.x, earth_to_stix.y, obstime=obstime)
    stix_to_earth = stix_coord.transform_to(Helioprojective(observer='Earth', obstime=obstime))

    assert_quantity_allclose(earth_to_stix.x, 410.747992 * u.arcsec)
    assert_quantity_allclose(earth_to_stix.y, 840.925692*u.arcsec)
    assert_quantity_allclose(earth_coord.Tx, stix_to_earth.Tx)
    assert_quantity_allclose(earth_coord.Ty, stix_to_earth.Ty)


@mock.patch('stixpy.frames._get_aux_data')
def test_hpc_to_stx(mock):
    # fake some data to make check easier
    obstime = Time('2021-05-22T02:52:00')
    roll, yaw, pitch, sas_x, sas_y = [0, 0, 0, 0, 0] * u.arcsec
    solo_pos = HeliographicStonyhurst(0*u.deg, 0*u.deg, 1*u.AU, obstime=obstime)
    mock.return_value = (roll, yaw, pitch, sas_x, sas_y, solo_pos.cartesian)
    solo_hpc_coord = Helioprojective(0*u.arcsec, 0*u.arcsec, observer=solo_pos, obstime=obstime)
    stix_frame = STIXImagingFrame(0*u.arcsec, 0*u.arcsec, obstime=obstime)
    stix_coord = solo_hpc_coord.transform_to(stix_frame)
    # should match the offset -8, 60
    assert_quantity_allclose(stix_coord.x, -8 * u.arcsec)
    assert_quantity_allclose(stix_coord.y, 60 * u.arcsec)


@mock.patch('stixpy.frames._get_aux_data')
def test_hpc_to_stx_no_sas(mock):
    # fake some data to make check easier
    obstime = Time('2021-05-22T02:52:00')
    roll, yaw, pitch, sas_x, sas_y = [0, 10, 10, np.nan, np.nan] * u.arcsec
    solo_pos = HeliographicStonyhurst(0*u.deg, 0*u.deg, 1*u.AU, obstime=obstime)
    mock.return_value = (roll, yaw, pitch, sas_x, sas_y, solo_pos.cartesian)
    solo_hpc_coord = Helioprojective(0*u.arcsec, 0*u.arcsec, observer=solo_pos, obstime=obstime)
    stix_frame = STIXImagingFrame(0*u.arcsec, 0*u.arcsec, obstime=obstime)
    with pytest.warns(Warning, match='SAS solution not'):
        stix_coord = solo_hpc_coord.transform_to(stix_frame)
        # should match the offset -8, 60 added to yaw and pitch 10, 10
        assert_quantity_allclose(stix_coord.x, (10-8) * u.arcsec)
        assert_quantity_allclose(stix_coord.y, (10+60) * u.arcsec)



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

import os
import sys

# sys.path.append(os.path.abspath('../'))


# -- Project information -----------------------------------------------------

project = 'STIXpy'
copyright = '2020, STIX Team'
author = 'STIX Team'

# The full version, including alpha/beta/rc tags
import stixpy
from stixpy import __version__
release = __version__
is_development = '.dev' in release

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.doctest',
    'sphinx.ext.mathjax',
    'sphinx_automodapi.automodapi',
    'sphinx_automodapi.smart_resolver',
    'sphinx_design',
    'sphinx_changelog',
    'matplotlib.sphinxext.plot_directive'
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '*.fits']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The reST default role (used for this markup: `text`) to use for all
# documents. Set to the "smart" one.
default_role = 'obj'

# Disable having a separate return type row
napoleon_use_rtype = False

# Disable google style docstrings
napoleon_google_docstring = False

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/',
               (None, 'http://data.astropy.org/intersphinx/python3.inv')),
    'numpy': ('https://docs.scipy.org/doc/numpy/',
              (None, 'http://data.astropy.org/intersphinx/numpy.inv')),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/',
              (None, 'http://data.astropy.org/intersphinx/scipy.inv')),
    'matplotlib': ('https://matplotlib.org/',
                   (None, 'http://data.astropy.org/intersphinx/matplotlib.inv')),
    'astropy': ('http://docs.astropy.org/en/stable/', None),
    'sunpy': ('https://docs.sunpy.org/en/stable/', None)}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

try:
    from sunpy_sphinx_theme.conf import *
except ImportError:
    html_theme = 'default'


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    '-Nfontsize=10',
    '-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Efontsize=10',
    '-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Gfontsize=10',
    '-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif'
]



import os
import sys
import tarfile
from pathlib import Path

import stixpy


package_dir = Path(os.path.dirname(stixpy.__file__))
root_dir = package_dir.joinpath("data")


def _unzip_test_data():
    data_tar = root_dir / 'test_data.tar.gz'
    with tarfile.open(data_tar.as_posix(), 'r:gz') as tar:
        def is_within_directory(directory, target):

            abs_directory = os.path.abspath(directory)
            abs_target = os.path.abspath(target)

            prefix = os.path.commonprefix([abs_directory, abs_target])

            return prefix == abs_directory

        def safe_extract(tar, path=".", members=None, *, numeric_owner=False):

            for member in tar.getmembers():
                member_path = os.path.join(path, member.name)
                if not is_within_directory(path, member_path):
                    raise Exception("Attempted Path Traversal in Tar File")

            tar.extractall(path, members, numeric_owner=numeric_owner)

        safe_extract(tar, root_dir.absolute())


_unzip_test_data()

_TEST_DATA = {
    'STIX_HK_MAXI_TIMESERIES': 'solo_L1_stix-hk-maxi_20200501_V01.fits',
    'STIX_QL_BACKGROUND_TIMESERIES': 'solo_L1_stix-ql-background_20200506_V01.fits',
    'STIX_QL_CALIBRATION': 'solo_L1_stix-cal-energy_20200506_V01.fits',
    'STIX_QL_FLAREFLAG':  'solo_L1_stix-ql-ql-tmstatusflarelist_20200506_V01.fits',
    'STIX_QL_LIGHTCURVE_TIMESERIES': 'solo_L1_stix-ql-lightcurve_20200506_V01.fits',
    'STIX_QL_SPECTRA': 'solo_L1_stix-ql-spectra_20200506_V01.fits',
    'STIX_QL_VARIANCE_TIMESERIES': 'solo_L1_stix-ql-variance_20200506_V01.fits',
    'STIX_SCI_XRAY_SPEC':
        'solo_L1_stix-sci-xray-spec_20200505T235959-20200506T000019_V01_0087031812-50886.fits',
    'STIX_SCI_XRAY_RPD':
        'solo_L1_stix-sci-xray-rpd_20200505T235959-20200506T000019_V01_0087031808-50882.fits',
    'STIX_SCI_XRAY_CPD':
        'solo_L1_stix-sci-xray-cpd_20200505T235959-20200506T000019_V01_0087031809-50883.fits',
    'STIX_SCI_XRAY_SCPD':
        'solo_L1_stix-sci-xray-scpd_20200505T235959-20200506T000019_V01_0087031810-50884.fits',
    # 'STIX_SCI_XRAY_VIS':
    #     'solo_L1_stix-sci-xray-vis-87031811_20200505T235958-20200510T000014_V01_50885.fits'
}

__doc__ = ''
for k, v in _TEST_DATA.items():
    p = root_dir / v
    if not p.exists():
        raise ValueError(f'Test data {k}, {v} missing please try manually running _unzip_test_date()')
    setattr(sys.modules[__name__], k, str(p))
    __doc__ = __doc__ + f'   - ``{k}``\n'

__all__ = [*_TEST_DATA.keys()]



from collections import defaultdict
from itertools import product

from pathlib import Path

import astropy.units as u
import numpy as np

from astropy.io import fits
from astropy.table import QTable, vstack
from astropy.time import Time, TimeDelta
from astropy.visualization import quantity_support
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.dates import date2num, HourLocator, DateFormatter
from matplotlib.widgets import Slider
from sunpy.time.timerange import TimeRange

__all__ = ['ScienceData', 'RawPixelData', 'CompressedPixelData', 'SummedCompressedPixelData',
           'Visibility', 'Spectrogram', 'TimesSeriesPlotMixin', 'SpectrogramPlotMixin',
           'PixelPlotMixin', 'PPrintMixin', 'IndexMasks', 'DetectorMasks', 'PixelMasks',
           'EnergyMasks']

from stixpy.product.product import L1Product

quantity_support()


class PPrintMixin:
    """
    Provides pretty printing for index masks.
    """
    @staticmethod
    def _pprint_indices(indices):
        groups = np.split(np.r_[:len(indices)], np.where(np.diff(indices) != 1)[0] + 1)
        out = ''
        for group in groups:
            if group.size < 3:
                out += f'{indices[group]}'
            else:
                out += f'[{indices[group[0]]}...{indices[group[-1]]}]'

        return out


class IndexMasks(PPrintMixin):
    """

    Attributes
    ----------
    masks : 'numpy.ndarray`
        The mask arrays
    indices : `numpy.ndarray'
        The indices the mask/s applies to

    """
    def __init__(self, mask_array):
        masks = np.unique(mask_array, axis=0)
        indices = [np.argwhere(np.all(mask_array == mask, axis=1)).reshape(-1)
                   for mask in masks]
        self.masks = masks
        self.indices = indices

    def __repr__(self):
        text = f'{self.__class__.__name__}\n'
        for m, i in zip(self.masks, self.indices):
            text += f'    {self._pprint_indices(i)}: ' \
                    f'[{",".join(np.where(m, np.arange(m.size), np.full(m.size, "_")))}]\n'
        return text


class DetectorMasks(IndexMasks):
    """
    Detector Index Masks
    """
    pass


class EnergyMasks(IndexMasks):
    """
    Energy Index Mask
    """
    pass


class PixelMasks(PPrintMixin):
    """
    Pixel Index Masks
    """
    def __init__(self, pixel_masks):
        masks = np.unique(pixel_masks, axis=0)
        indices = []
        if masks.ndim == 2:
            indices = [np.argwhere(np.all(pixel_masks == mask, axis=1)).reshape(-1)
                       for mask in masks]
        elif masks.ndim == 3:
            indices = [np.argwhere(np.all(pixel_masks == mask, axis=(1, 2))).reshape(-1)
                       for mask in masks]
        self.masks = masks
        self.indices = indices

    def __repr__(self):
        text = f'{self.__class__.__name__}\n'
        for m, i in zip(self.masks, self.indices):
            text += f'    {self._pprint_indices(i)}: ' \
                    f'[{str(np.where(m.shape[0], m, np.full(m.shape, "_")))}]\n'
        return text


class SpectrogramPlotMixin:
    """
    Spectrogram plot mixin providing spectrogram plotting for pixel data.
    """
    def plot_spectrogram(self, axes=None, time_indices=None, energy_indices=None,
                         detector_indices='all', pixel_indices='all', **plot_kwargs):
        """
        Plot a spectrogram for the selected time and energies.

        Parameters
        ----------
        axes : optional `matplotlib.axes`
            The axes the plot the spectrogram.
        time_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `time_indices=[0, 2, 5]` would return only the first, third and
            sixth times while `time_indices=[[0, 2],[3, 5]]` would sum the data between.
        energy_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `energy_indices=[0, 2, 5]` would return only the first, third and
            sixth times while `energy_indices=[[0, 2],[3, 5]]` would sum the data between.
        **plot_kwargs : `dict`
            Any additional arguments are passed to :meth:`~matplotlib.axes.Axes.pcolormesh`.

        Returns
        -------
        `matplotlib.axes`
        """
        counts_shape = self.data['counts'].shape
        if len(counts_shape) != 4:
            # if spectrogram can't do anything with pixel or detector indices
            if detector_indices != 'all' or pixel_indices != 'all':
                raise ValueError('Detector and or pixel indices have can not be used with spectrogram')

            pid = None
            did = None
        else:

            if detector_indices == 'all':
                did = [[0, 31]]
            else:
                det_idx_arr = np.array(detector_indices)
                if det_idx_arr.ndim == 1 and det_idx_arr.size != 1:
                    raise ValueError('Spectrogram plots can only show data from a single '
                                     'detector or summed over a number of detectors')
                elif det_idx_arr.ndim == 2 and det_idx_arr.shape[0] != 1:
                    raise ValueError('Spectrogram plots can only one sum '
                                     'detector or summed over a number of detectors')
                did = detector_indices

            if pixel_indices == 'all':
                pid = [[0, 11]]
            else:
                pix_idx_arr = np.array(pixel_indices)
                if pix_idx_arr.ndim == 1 and pix_idx_arr.size != 1:
                    raise ValueError('Spectrogram plots can only show data from a single '
                                     'detector or summed over a number of detectors')
                elif pix_idx_arr.ndim == 2 and pix_idx_arr.shape[0] != 1:
                    raise ValueError('Spectrogram plots can only one sum '
                                     'detector or summed over a number of detectors')
                pid = pixel_indices

        counts, errors, times, timedeltas, energies = self.get_data(detector_indices=did,
                                                                    pixel_indices=pid,
                                                                    time_indices=time_indices,
                                                                    energy_indices=energy_indices)

        e_edges = np.hstack([energies['e_low'], energies['e_high'][-1]])
        t_edges = Time(np.concatenate(
            [times - timedeltas.reshape(-1) / 2, times[-1] + timedeltas.reshape(-1)[-1:] / 2]))

        if axes is None:
            fig, axes = plt.subplots()
        else:
            fig = axes.get_figure()

        pcolor_kwargs = {'norm':LogNorm(), 'shading':'flat'}
        pcolor_kwargs.update(plot_kwargs)
        im = axes.pcolormesh(t_edges.datetime, e_edges[1:-1], counts[:, 0, 0, 1:-1].T.value,
                             **pcolor_kwargs)

        #axes.colorbar(im).set_label(format(counts.unit))
        axes.xaxis_date()
        # axes.set_yticks(range(y_lims[0], y_lims[1] + 1))
        # axes.set_yticklabels(labels)
        minor_loc = HourLocator()
        axes.xaxis.set_minor_locator(minor_loc)
        axes.xaxis.set_major_formatter(DateFormatter("%d %H:%M"))
        # fig.autofmt_xdate()
        # fig.tight_layout()

        return axes


class TimesSeriesPlotMixin:
    """
    TimesSeries plot mixin providing timeseries plotting for pixel data.
    """
    def plot_timeseries(self, time_indices=None, energy_indices=None, detector_indices='all',
                        pixel_indices='all', axes=None, error_bar=False, **plot_kwarg):
        """
        Plot a times series of the selected times and energies.

        Parameters
        ----------
        axes : optional `matplotlib.axes`
            The axes the plot the spectrogram.
        time_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `time_indices=[0, 2, 5]` would return only the first, third and
            sixth times while `time_indices=[[0, 2],[3, 5]]` would sum the data between.
        energy_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `energy_indices=[0, 2, 5]` would return only the first, third and
            sixth times while `energy_indices=[[0, 2],[3, 5]]` would sum the data between.
        detector_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `detector_indices=[0, 2, 5]` would return only the first, third and
            sixth detectors while `detector_indices=[[0, 2],[3, 5]]` would sum the data between.
        pixel_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `pixel_indices=[0, 2, 5]` would return only the first, third and
            sixth pixels while `pixel_indices=[[0, 2],[3, 5]]` would sum the data between.
        error_bar : optional `bool`
            Add error bars to plot.
        **plot_kwargs : `dict`
            Any additional arguments are passed to :meth:`~matplotlib.axes.Axes.plot`.

        Returns
        -------
        `matplotlib.axes`

        """
        if axes is None:
            fig, axes = plt.subplots()
        else:
            fig = axes.get_figure()

        if detector_indices == 'all':
            detector_indices = [[0, 31]]

        if pixel_indices == 'all':
            pixel_indices = [[0, 11]]

        counts, errors, times, timedeltas, energies \
            = self.get_data(detector_indices=detector_indices,
                            pixel_indices=pixel_indices,
                            time_indices=time_indices, energy_indices=energy_indices)

        labels = [f'{el.value} - {eh.value}' for el, eh in energies['e_low', 'e_high']]

        n_time, n_det, n_pix, n_energy = counts.shape

        for did, pid, eid in product(range(n_det), range(n_pix), range(n_energy)):
            if error_bar:
                lines = axes.errorbar(times.to_datetime(), counts[:, did, pid, eid], yerr=errors[:, did, pid, eid],
                                      label=labels[eid], **plot_kwarg)
            else:
                lines = axes.plot(times.to_datetime(), counts[:, did, pid, eid],
                                  label=labels[eid], **plot_kwarg)

        axes.set_yscale('log')
        axes.xaxis.set_major_formatter(DateFormatter("%d %H:%M"))
        fig.autofmt_xdate()
        fig.tight_layout()

        return lines


class PixelPlotMixin:
    """
    Pixel plot mixin providing pixel plotting for pixel data.
    """
    def plot_pixels(self, time_indices=None, energy_indices=None, fig=None):
        """
        Plot individual pixel data for each detector.

        Parameters
        ----------
        time_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `time_indices=[0, 2, 5]` would return only the first, third and
            sixth times while `time_indices=[[0, 2],[3, 5]]` would sum the data between.
        energy_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `energy_indices=[0, 2, 5]` would return only the first, third and
            sixth times while `energy_indices=[[0, 2],[3, 5]]` would sum the data between.
        fig : optional `matplotlib.figure`
            The figure where to which the pixel plot will be added.

        Returns
        -------
        `matplotlib.figure`
            The figure
        """

        if fig:
            axes = fig.subplots(nrows=4, ncols=8, sharex=True, sharey=True, figsize=(10, 5))
        else:
            fig, axes = plt.subplots(nrows=4, ncols=8, sharex=True, sharey=True, figsize=(10, 5))

        counts, count_err, times, dt, energies = self.get_data(time_indices=time_indices,
                                                               energy_indices=energy_indices)

        def timeval(val):
            return times[val].isot

        def energyval(val):
            return f"{energies[val]['e_low'].value}-{energies[val]['e_high']}"

        axcolor = 'lightgoldenrodyellow'
        axenergy = plt.axes([0.15, 0.05, 0.55, 0.03], facecolor=axcolor)
        senergy = SliderCustomValue(axenergy, 'Energy', 0, len(energies)-1, format_func=energyval,
                                    valinit=0, valstep=1)
        axetime = plt.axes([0.15, 0.01, 0.55, 0.03], facecolor=axcolor)
        stime = SliderCustomValue(axetime, 'Time', 0, counts.shape[0]-1, format_func=timeval,
                                  valinit=0, valstep=1)

        pixel_ids = [slice(0, 4), slice(4, 8), slice(8, 12)]
        if counts.shape[2] == 4:
            pixel_ids = [slice(0, 4)]

        containers = defaultdict(list)

        for detector_id in range(32):
            row, col = divmod(detector_id, 8)
            for pixel_id in pixel_ids:
                errbar_cont = axes[row, col].errorbar((0.5, 1.5, 2.5, 3.5),
                                                      counts[0, detector_id, pixel_id, 0],
                                                      yerr=count_err[0, detector_id, pixel_id, 0],
                                                      xerr=0.5, ls='')

                containers[row, col].append(errbar_cont)
                axes[row, col].set_xlim(0, 4)
                axes[row, col].set_title(f'Det {detector_id}')
                axes[row, col].set_xticks([])
                if col != 0:
                    # axs[row, col].set_yticks([])
                    axes[row, col].set_ylabel('')

        def update(_):
            energy_index = senergy.val
            time_index = stime.val
            pids_ = [slice(0, 4), slice(4, 8), slice(8, 12)]
            if counts.shape[2] == 4:
                pids_ = [slice(0, 4)]

            imaging_mask = np.ones(32, bool)
            imaging_mask[8:10] = False

            for did in range(32):
                r, c = divmod(did, 8)
                axes[r, c].set_ylim(0, counts[time_index, imaging_mask, :,
                                              energy_index].max()*1.2)

                for i, pid in enumerate(pids_):
                    lines, caps, bars = containers[r, c][i]
                    lines.set_ydata(counts[time_index, did, pid, energy_index])

                    # horizontal bars at value
                    segs = np.array(bars[0].get_segments())
                    segs[:, 0, 0] = [0., 1., 2., 3.]
                    segs[:, 1, 0] = [1., 2., 3., 4.]
                    segs[:, 0, 1] = counts[time_index, did, pid, energy_index]
                    segs[:, 1, 1] = counts[time_index, did, pid, energy_index]
                    bars[0].set_segments(segs)
                    # vertical bars at +/- error
                    segs = np.array(bars[1].get_segments())
                    segs[:, 0, 0] = [0.5, 1.5, 2.5, 3.5]
                    segs[:, 1, 0] = [0.5, 1.5, 2.5, 3.5]
                    segs[:, 0, 1] = counts[time_index, did, pid, energy_index] \
                        - count_err[time_index, did, pid, energy_index]
                    segs[:, 1, 1] = counts[time_index, did, pid, energy_index] \
                        + count_err[time_index, did, pid, energy_index]
                    bars[1].set_segments(segs)

        senergy.on_changed(update)
        stime.on_changed(update)


class ScienceData(L1Product):
    """
    Basic science data class
    """
    def __init__(self, *, meta, control, data, energies, idb_versions=None):
        """

        Parameters
        ----------
        header : `astropy.fits.Header`
            Fits header
        control : `astropy.table.QTable`
            Fits file control extension
        data :` astropy.table.QTable`
            Fits file data extension
        energies : `astropy.table.QTable`
            Fits file energy extension
        """
        super().__init__(meta=meta, control=control, data=data,
                         energies=energies, idb_versions=idb_versions)

        self.count_type = 'rate'
        if 'detector_masks' in self.data.colnames:
            self.detector_masks = DetectorMasks(self.data['detector_masks'])
        if 'pixel_masks' in self.data.colnames:
            self.pixel_masks = PixelMasks(self.data['pixel_masks'])
        if 'energy_bin_edge_mask' in self.control.colnames:
            self.energy_masks = EnergyMasks(self.control['energy_bin_edge_mask'])
            self.dE = energies['e_high'] - energies['e_low']

    @property
    def time_range(self):
        """
        A `sunpy.time.TimeRange` for the data.
        """
        return TimeRange(self.data['time'][0] - self.data['timedel'][0]/2,
                         self.data['time'][-1] + self.data['timedel'][-1]/2)

    @property
    def pixels(self):
        """
        A `stixpy.science.PixelMasks` object representing the pixels contained in the data
        """
        return self.pixel_masks

    @property
    def detectors(self):
        """
        A `stixpy.science.DetectorMasks` object representing the detectors contained in the data.
        """
        return self.detector_masks

    @property
    def energies(self):
        """
        A `astropy.table.Table` object representing the energies contained in the data.
        """
        return self._energies

    @property
    def times(self):
        """
        An `astropy.time.Time` array representing the center of the observed time bins.
        """
        return self.data['time']

    @property
    def durtaion(self):
        """
        An `astropy.units.Quantiy` array giving the duration or integration time
        """
        return self.data['timedel']

    def get_data(self, time_indices=None, energy_indices=None, detector_indices=None,
                 pixel_indices=None, sum_all_times=False):
        """
        Return the counts, errors, times, durations and energies for selected data.

        Optionally summing in time and or energy.

        Parameters
        ----------
        time_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `time_indices=[0, 2, 5]` would return only the first, third and
            sixth times while `time_indices=[[0, 2],[3, 5]]` would sum the data between.
        energy_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `energy_indices=[0, 2, 5]` would return only the first, third and
            sixth times while `energy_indices=[[0, 2],[3, 5]]` would sum the data between.
        detector_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `detector_indices=[0, 2, 5]` would return only the first, third and
            sixth detectors while `detector_indices=[[0, 2],[3, 5]]` would sum the data between.
        pixel_indices : `list` or `numpy.ndarray`
            If an 1xN array will be treated as mask if 2XN array will sum data between given
            indices. For example `pixel_indices=[0, 2, 5]` would return only the first, third and
            sixth pixels while `pixel_indices=[[0, 2],[3, 5]]` would sum the data between.
        sum_all_times : `bool`
            Flag to sum all give time intervals into one

        Returns
        -------
        `tuple`
            Counts, errors, times, deltatimes,  energies

        """
        counts = self.data['counts']
        try:
            counts_var = self.data['counts_comp_err']**2
        except KeyError:
            counts_var = self.data['counts_comp_comp_err'] ** 2
        shape = counts.shape
        if len(shape) < 4:
            counts = counts.reshape(shape[0], 1, 1, shape[-1])
            counts_var = counts_var.reshape(shape[0], 1, 1, shape[-1])

        energies = self.energies[:]
        times = self.times

        if detector_indices is not None:
            detecor_indices = np.asarray(detector_indices)
            if detecor_indices.ndim == 1:
                detector_mask = np.full(32, False)
                detector_mask[detecor_indices] = True
                counts = counts[:, detector_mask, ...]
                counts_var = counts_var[:, detector_mask, ...]
            elif detecor_indices.ndim == 2:
                counts = np.hstack([np.sum(counts[:, dl:dh + 1, ...], axis=1, keepdims=True)
                                    for dl, dh in detecor_indices])

                counts_var = np.concatenate(
                    [np.sum(counts_var[:, dl:dh + 1, ...], axis=1, keepdims=True)
                     for dl, dh in detecor_indices], axis=1)

        if pixel_indices is not None:
            pixel_indices = np.asarray(pixel_indices)
            if pixel_indices.ndim == 1:
                pixel_mask = np.full(12, False)
                pixel_mask[pixel_indices] = True
                counts = counts[..., pixel_mask, :]
                counts_var = counts_var[..., pixel_mask, :]
            elif pixel_indices.ndim == 2:
                counts = np.concatenate([np.sum(counts[..., pl:ph + 1, :], axis=2, keepdims=True)
                                         for pl, ph in pixel_indices], axis=2)

                counts_var = np.concatenate(
                    [np.sum(counts_var[..., pl:ph + 1, :], axis=2, keepdims=True)
                     for pl, ph in pixel_indices], axis=2)

        e_norm = self.dE
        if energy_indices is not None:
            energy_indices = np.asarray(energy_indices)
            if energy_indices.ndim == 1:
                energy_mask = np.full(shape[-1], False)
                energy_mask[energy_indices] = True
                counts = counts[..., energy_mask]
                counts_var = counts_var[..., energy_mask]
                e_norm = self.dE[energy_mask]
                energies = self.energies[energy_mask]
            elif energy_indices.ndim == 2:
                counts = np.concatenate([np.sum(counts[..., el:eh+1], axis=-1, keepdims=True)
                                         for el, eh in energy_indices], axis=-1)

                counts_var = np.concatenate(
                    [np.sum(counts_var[..., el:eh+1], axis=-1, keepdims=True)
                     for el, eh in energy_indices], axis=-1)

                e_norm = np.hstack([(energies['e_high'][eh] - energies['e_low'][el])
                                    for el, eh in energy_indices])

                energies = np.atleast_2d(
                    [(self._energies['e_low'][el].value, self._energies['e_high'][eh].value)
                     for el, eh in energy_indices])
                energies = QTable(energies*u.keV, names=['e_low', 'e_high'])

        t_norm = self.data['timedel']
        if time_indices is not None:
            time_indices = np.asarray(time_indices)
            if time_indices.ndim == 1:
                time_mask = np.full(times.shape, False)
                time_mask[time_indices] = True
                counts = counts[time_mask, ...]
                counts_var = counts_var[time_mask, ...]
                t_norm = self.data['timedel'][time_mask]
                times = times[time_mask]
                # dT = self.data['timedel'][time_mask]
            elif time_indices.ndim == 2:
                new_times = []
                dt = []
                for tl, th in time_indices:
                    ts = times[tl] - self.data['timedel'][tl] * 0.5
                    te = times[th] + self.data['timedel'][th] * 0.5
                    td = te - ts
                    tc = ts + (td * 0.5)
                    dt.append(td.to('s'))
                    new_times.append(tc)
                dt = np.hstack(dt)
                times = Time(new_times)
                counts = np.vstack([np.sum(counts[tl:th+1, ...], axis=0, keepdims=True) for tl, th
                                    in time_indices])

                counts_var = np.vstack([np.sum(counts_var[tl:th+1, ...], axis=0, keepdims=True)
                                        for tl, th in time_indices])
                t_norm = dt

                if sum_all_times and len(new_times) > 1:
                    counts = np.sum(counts, axis=0, keepdims=True)
                    counts_var = np.sum(counts_var, axis=0, keepdims=True)
                    t_norm = np.sum(dt)

        if e_norm.size != 1:
            e_norm = e_norm.reshape(1, 1, 1, -1)

        if t_norm.size != 1:
            t_norm = t_norm.reshape(-1, 1, 1, 1)

        counts_err = np.sqrt(counts*u.ct + counts_var)/(e_norm * t_norm)
        counts = counts / (e_norm * t_norm)

        return counts, counts_err, times, t_norm, energies

    def concatenate(self, others):
        """
        Concatenate two or more science products.

        Parameters
        ----------
        others: `list` [`stixpy.science.ScienceData`]
            The other/s science products to concatenate

        Returns
        -------
        `stixpy.science.ScienceData`
            The concatenated science products
        """
        others = others if isinstance(others, list) else [others]
        if all([isinstance(o, type(self)) for o in others]):
            control = self.control[:]
            data = self.data[:]
            for other in others:
                self_control_ind_max = data['control_index'].max() + 1
                other.control['index'] = other.control['index'] + self_control_ind_max
                other.data['control_index'] = other.data['control_index'] + self_control_ind_max

                try:
                    [(table.meta.pop('DATASUM'), table.meta.pop('CHECKSUM'))
                     for table in [control, other.control, data, other.data]]
                except KeyError:
                    pass

                control = vstack([control, other.control])
                data = vstack([data, other.data])

            return type(self)(meta=self.meta, control=control, data=data,
                              energies=self.energies, idb_version=self.idb_versions)

    def __repr__(self):
        return f'{self.__class__.__name__}' \
               f'{self.time_range}' \
               f'    {self.detector_masks}\n' \
               f'    {self.pixel_masks}\n' \
               f'    {self.energy_masks}'


class RawPixelData(ScienceData, PixelPlotMixin, TimesSeriesPlotMixin, SpectrogramPlotMixin):
    """
    Uncompressed or raw count data from selected pixels, detectors and energies.

    Examples
    --------
    >>> from stixpy.data import test
    >>> from stixpy.science import ScienceData
    >>> raw_pd = ScienceData.from_fits(test.STIX_SCI_XRAY_RPD)
    >>> raw_pd
    RawPixelData   <sunpy.time.timerange.TimeRange object at ...
        Start: 2020-05-05 23:59:59
        End:   2020-05-06 00:00:19
        Center:2020-05-06 00:00:09
    Duration:0.00023148148148144365 days or
               0.005555555555554648 hours or
               0.33333333333327886 minutes or
               19.99999999999673 seconds
        DetectorMasks
        [0...4]: [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]
    <BLANKLINE>
        PixelMasks
        [0...4]: [['1' '1' '1' '1' '1' '1' '1' '1' '1' '1' '1' '1']]
    <BLANKLINE>
        EnergyMasks
        [0]: [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]
    <BLANKLINE>
    """
    pass

    @classmethod
    def is_datasource_for(cls, *, meta, **kwargs):
        """Determines if meta data meach Raw Pixel Data"""
        service_subservice_ssid = tuple(meta[name] for name in ['STYPE', 'SSTYPE', 'SSID'])
        level = meta['level']
        if service_subservice_ssid == (21, 6, 20) and level == 'L1':
            return True


class CompressedPixelData(ScienceData, PixelPlotMixin, TimesSeriesPlotMixin, SpectrogramPlotMixin):
    """
    Compressed count data from selected pixels, detectors and energies.

    Examples
    --------
    >>> from stixpy.data import test
    >>> from stixpy.science import ScienceData
    >>> compressed_pd = ScienceData.from_fits(test.STIX_SCI_XRAY_CPD)
    >>> compressed_pd
    CompressedPixelData   <sunpy.time.timerange.TimeRange object at ...
        Start: 2020-05-05 23:59:59
        End:   2020-05-06 00:00:19
        Center:2020-05-06 00:00:09
        Duration:0.00023148148148144365 days or
               0.005555555555554648 hours or
               0.33333333333327886 minutes or
               19.99999999999673 seconds
        DetectorMasks
        [0...4]: [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]
    <BLANKLINE>
        PixelMasks
        [0...4]: [['1' '1' '1' '1' '1' '1' '1' '1' '1' '1' '1' '1']]
    <BLANKLINE>
        EnergyMasks
        [0]: [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]
    """

    @classmethod
    def is_datasource_for(cls, *, meta, **kwargs):
        """Determines if meta data meach Raw Pixel Data"""
        service_subservice_ssid = tuple(meta[name] for name in ['STYPE', 'SSTYPE', 'SSID'])
        level = meta['level']
        if service_subservice_ssid == (21, 6, 21) and level == 'L1':
            return True


class SummedCompressedPixelData(ScienceData, PixelPlotMixin, TimesSeriesPlotMixin,
                                SpectrogramPlotMixin):
    """
    Compressed and Summed count data from selected pixels, detectors and energies.

    Examples
    --------
    >>> from stixpy.data import test
    >>> from stixpy.science import ScienceData
    >>> summed_pd = ScienceData.from_fits(test.STIX_SCI_XRAY_SCPD)
    >>> summed_pd
    SummedCompressedPixelData   <sunpy.time.timerange.TimeRange object at ...
        Start: 2020-05-05 23:59:59
        End:   2020-05-06 00:00:19
        Center:2020-05-06 00:00:09
        Duration:0.00023148148148144365 days or
               0.005555555555554648 hours or
               0.33333333333327886 minutes or
               19.99999999999673 seconds
        DetectorMasks
        [0...4]: [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]
    <BLANKLINE>
        PixelMasks
        [0...4]: [[['0' '0' '0' '1' '0' '0' '0' '1' '0' '0' '0' '1']
     ['0' '0' '1' '0' '0' '0' '1' '0' '0' '0' '1' '0']
     ['0' '1' '0' '0' '0' '1' '0' '0' '0' '1' '0' '0']
     ['1' '0' '0' '0' '1' '0' '0' '0' '1' '0' '0' '0']]]
    <BLANKLINE>
        EnergyMasks
        [0]: [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]
    """
    pass

    @classmethod
    def is_datasource_for(cls, *, meta, **kwargs):
        """Determines if meta data meach Raw Pixel Data"""
        service_subservice_ssid = tuple(meta[name] for name in ['STYPE', 'SSTYPE', 'SSID'])
        level = meta['level']
        if service_subservice_ssid == (21, 6, 22) and level == 'L1':
            return True


class Visibility(ScienceData):
    """
    Compressed visibilities from selected pixels, detectors and energies.

    Examples
    --------
    # >>> from stixpy.data import test
    # >>> from stixpy.science import ScienceData
    # >>> visibility = ScienceData.from_fits(test.STIX_SCI_XRAY_VIZ)
    # >>> visibility
    # Visibility   <sunpy.time.timerange.TimeRange object at ...>
    #     Start: 2020-05-07 23:59:58
    #     End:   2020-05-08 00:00:14
    #     Center:2020-05-08 00:00:06
    #     Duration:0.00018518518518517713 days or
    #            0.004444444444444251 hours or
    #            0.26666666666665506 minutes or
    #            15.999999999999304 seconds
    #     DetectorMasks
    #     [0...4]: [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]
    # <BLANKLINE>
    #     PixelMasks
    #     [0]: [[['1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0']]]
    #     [1]: [[['1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0']]]
    #     [2]: [[['1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0']]]
    #     [3]: [[['1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0']]]
    #     [4]: [[['1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0' '0.0']
    #  ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '1.0']]]
    # <BLANKLINE>
    #     EnergyMasks
    #     [0]: [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]
    # <BLANKLINE>
    """
    def __init__(self, *, header, control, data, energies):
        super().__init__(meta=header, control=control, data=data, energies=energies)
        self.pixel_masks = PixelMasks(self.pixels)

    @property
    def pixels(self):
        return np.vstack([self.data[f'pixel_mask{i}'][0] for i in range(1, 6)])

    @classmethod
    def is_datasource_for(cls, *, meta, **kwargs):
        """Determines if meta data meach Raw Pixel Data"""
        service_subservice_ssid = tuple(meta[name] for name in ['STYPE', 'SSTYPE', 'SSID'])
        level = meta['level']
        if service_subservice_ssid == (21, 6, 23) and level == 'L1':
            return True


class Spectrogram(ScienceData, TimesSeriesPlotMixin, SpectrogramPlotMixin):
    """
    Spectrogram from selected pixels, detectors and energies.

    Parameters
    ----------
    header : `astropy.fits.Header`
    control : `astropy.table.QTable`
    data : `astropy.table.QTable`
    energies : `astropy.table.QTable`

    Examples
    --------
    >>> from stixpy.data import test
    >>> from stixpy.science import ScienceData
    >>> spectogram = ScienceData.from_fits(test.STIX_SCI_XRAY_SPEC)
    >>> spectogram
    Spectrogram   <sunpy.time.timerange.TimeRange ...
        Start: 2020-05-05 23:59:59
        End:   2020-05-06 00:00:19
        Center:2020-05-06 00:00:09
        Duration:0.00023148148148144365 days or
            0.005555555555554648 hours or
            0.33333333333327886 minutes or
            19.99999999999673 seconds
        DetectorMasks
        [0]: [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]
    <BLANKLINE>
        PixelMasks
        [0...4]: [['0' '0' '0' '0' '0' '0' '0' '0' '0' '0' '0' '0']]
    <BLANKLINE>
        EnergyMasks
        [0]: [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]
    <BLANKLINE>
    """
    def __init__(self, *, meta, control, data, energies, idb_versions):
        """

        Parameters
        ----------
        header : astropy.fits.Header
        control : astropy.table.QTable
        data : astropy.table.QTable
        energies : astropy.table.QTable
        """
        super().__init__(meta=meta, control=control, data=data,
                         energies=energies, idb_versions=idb_versions)
        self.count_type = 'rate'
        self.detector_masks = DetectorMasks(self.control['detector_masks'])
        self.pixel_masks = PixelMasks(self.data['pixel_masks'])
        self.energy_masks = EnergyMasks(self.control['energy_bin_mask'])
        self.dE = (energies['e_high'] - energies['e_low'])[self.energy_masks.masks[0] == 1]
        # self.dE = np.hstack([[1], np.diff(energies['e_low'][1:]).value, [1]]) * u.keV

    @classmethod
    def is_datasource_for(cls, *, meta, **kwargs):
        """Determines if meta data meach Raw Pixel Data"""
        service_subservice_ssid = tuple(meta[name] for name in ['STYPE', 'SSTYPE', 'SSID'])
        level = meta['level']
        if service_subservice_ssid == (21, 6, 24) and level == 'L1':
            return True


class SliderCustomValue(Slider):
    """
    A slider with a customisable formatter
    """
    def __init__(self, *args, format_func=None, **kwargs):
        if format_func is not None:
            self._format = format_func
        super().__init__(*args, **kwargs)



import pytest
from sunpy.net import Fido, attrs as a


from stixpy.net.client import STIXClient


@pytest.fixture
def client():
    return STIXClient()


@pytest.mark.remote_data
def test_search_date(client):
    res = client.search(a.Time('2020-05-01T00:00', '2020-05-01T23:59'), a.Instrument.stix)
    assert len(res) == 37


@pytest.mark.remote_data
def test_search_date_product():
    res1 = Fido.search(a.Time('2020-05-01T00:00', '2020-05-01T23:59'),
                      a.stix.DataProduct.ql_lightcurve, a.Instrument.stix)
    assert len(res1[0]) == 1

    res2 = Fido.search(a.Time('2020-05-01T00:00', '2020-05-02T00:00'),
                      a.stix.DataProduct.ql_lightcurve, a.Instrument.stix)
    assert len(res2[0]) == 2

    res3 = Fido.search(a.Time("2022-01-20 05:40", "2022-01-20 06:20"), a.Instrument("STIX"),
                      a.stix.DataType("sci"))
    assert len(res3[0]) == 9


@pytest.mark.remote_data
def test_search_date_product_xray_level1():
    res = Fido.search(a.Time('2020-05-23T00:00', '2020-05-25T00:00'), a.Instrument.stix,
                      a.stix.DataProduct.sci_xray_l1, a.stix.DataType.sci)
    assert len(res) == 1


@pytest.mark.remote_data
def test_search_date_product_xray_level1():
    res = Fido.search(a.Time('2021-05-07T00:00', '2021-05-08T00:00'), a.Instrument.stix,
                      a.stix.DataType.sci)
    assert len(res) == 1


@pytest.mark.remote_data
def test_fido():
    res = Fido.search(a.Time('2020-11-17T00:00', '2020-11-17T23:59'), a.Instrument.stix)
    assert len(res['stix']) == 50

    res_ql = Fido.search(a.Time('2020-11-17T00:00', '2020-11-17T23:59'), a.Instrument.stix,
                         a.stix.DataType.ql)
    assert len(res_ql['stix']) == 5

    res_sci = Fido.search(a.Time('2020-11-17T00:00', '2020-11-17T23:59'), a.Instrument.stix,
                          a.stix.DataType.sci)
    assert len(res_sci['stix']) == 42



"""
The logging module provides a common logging using pythons logging module. A root logger is
configured for the stixcore package on import and all subsequent loggers should be created as
children of the logger.

Examples
--------
.. code-block:: python

   #stixcore.mymodule (stixcore/mymodule.py)
   import logging
   from stixcore.util.logging import get_logger

   logger = get_logger(__name__)
   logger.debug('you will not see me')
   logger = get_logger(__name__, level=logging.DEBUG)
   logger.debug('you will see me')

   2020-11-25T12:21:27Z DEBUG: you will see me - stixcore.mymodule

"""
import logging

STX_LOGGER_FORMAT = '%(asctime)s %(levelname)s %(name)s %(lineno)s: %(message)s'
STX_LOGGER_DATE_FORMAT = '%Y-%m-%dT%H:%M:%SZ'


def get_logger(name, level=logging.INFO):
    """
    Return a configured logger instance.

    Parameters
    ----------
    name : `str`
        Name of the logger
    level : `int` or level, optional
        Level of the logger e.g `logging.DEBUG`

    Returns
    -------
    `logging.Logger`
        Configured logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(STX_LOGGER_FORMAT, datefmt=STX_LOGGER_DATE_FORMAT)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



import astropy.units as u
import numpy as np

from stixpy.data import test
from stixpy.product.sources import CompressedPixelData, SummedCompressedPixelData, Spectrogram, RawPixelData
from stixpy.science import ScienceData


def test_sciencedata_get_data():
    l1 = ScienceData.from_fits(test.STIX_SCI_XRAY_CPD)
    tot = l1.data['counts']
    norm = (l1.data['timedel'].reshape(5, 1, 1, 1) * l1.dE)
    rate = tot / norm
    error = np.sqrt(tot*u.ct+l1.data['counts_comp_err']**2) / norm
    r, re, t, dt, e = l1.get_data()
    assert np.allclose(rate, r)
    assert np.allclose(error, re)

    # Detector sum
    tot = l1.data['counts'][:, 0:32, ...].sum(axis=1, keepdims=True)
    norm = (l1.data['timedel'].reshape(5, 1, 1, 1) * l1.dE)
    rate = tot / norm
    error = np.sqrt(tot*u.ct+l1.data['counts_comp_err'][:, 0:32, ...].sum(axis=1, keepdims=True)**2)/norm
    r, re, t, dt, e = l1.get_data(detector_indices=[[0, 31]])
    assert np.allclose(rate, r)
    assert np.allclose(error, re, atol=1e-3)

    # Pixel sum
    tot = l1.data['counts'][..., 0:12, :].sum(axis=2, keepdims=True)
    norm = (l1.data['timedel'].reshape(5, 1, 1, 1) * l1.dE)
    rate = tot / norm
    error = np.sqrt(tot * u.ct
                    + l1.data['counts_comp_err'][..., 0:12, :].sum(axis=2, keepdims=True)**2) / norm
    r, re, t, dt, e = l1.get_data(pixel_indices=[[0, 11]])
    assert np.allclose(rate, r)
    assert np.allclose(error, re)

    # Detector and Pixel sum
    tot = l1.data['counts'][:, 0:32, 0:12, :].sum(axis=(1, 2), keepdims=True)
    norm = (l1.data['timedel'].reshape(5, 1, 1, 1) * l1.dE)
    rate = tot / norm
    error = np.sqrt(tot*u.ct + l1.data['counts_comp_err'][:, 0:32, 0:12, :].sum(axis=(1, 2),
                                                                           keepdims=True)**2) / norm
    r, re, t, dt, e = l1.get_data(pixel_indices=[[0, 11]], detector_indices=[[0, 31]])
    assert np.allclose(rate, r)
    assert np.allclose(error, re, atol=1e-3)

    # Energy sum
    tot = l1.data['counts'][..., 1:31].sum(axis=3, keepdims=True)
    norm = (l1.data['timedel'].reshape(5, 1, 1, 1)
            * (l1.energies[30]['e_high']-l1.energies[1]['e_low']))
    rate = tot / norm
    error = np.sqrt(tot*u.ct + l1.data['counts_comp_err'][..., 1:31].sum(axis=3, keepdims=True)**2)/norm
    r, re, t, dt, e = l1.get_data(energy_indices=[[1, 30]])
    assert np.allclose(rate, r)
    assert np.allclose(error, re, atol=1e-3)

    # Time sum
    tot = l1.data['counts'][:, ...].sum(axis=0, keepdims=True)
    norm = (l1.data['timedel'].sum() * l1.dE)
    rate = tot / norm
    error = np.sqrt(tot * u.ct + l1.data['counts_comp_err'][:, ...].sum(axis=0, keepdims=True) ** 2)/norm
    r, re, t, dt, e = l1.get_data(time_indices=[[0, 4]])
    assert np.allclose(rate, r)
    assert np.allclose(error, re)

    # Sum everything down to one number
    tot = l1.data['counts'][..., 1:31].sum(keepdims=True)
    norm = (l1.data['timedel'].sum() * (l1.energies[30]['e_high'] - l1.energies[1]['e_low']))
    rate = tot/norm
    error = np.sqrt(tot * u.ct + l1.data['counts_comp_err'][..., 1:31].sum(keepdims=True) ** 2) / norm
    r, re, t, dt, e = l1.get_data(time_indices=[[0, 4]], energy_indices=[[1, 30]],
                                  pixel_indices=[[0, 11]], detector_indices=[[0, 31]])
    assert np.allclose(rate, r)
    assert np.allclose(error, re, atol=1e-3)

    # Overwrite data with known values
    # 5 seconds, with dE 1/30keV across 30 channels
    l1.data.remove_column('counts') # need to remove and add to avoid error regarding casting
    l1.data['counts'] = np.full((5, 32, 12, 32), 146/(5*30))*u.ct
    l1.data.remove_column('counts_comp_err')
    l1.data['counts_comp_err'] = np.full_like(l1.data['counts'], np.sqrt(146*145/(5*30))*u.ct)
    l1.data['timedel'] = 1/5*u.s
    l1.dE[:] = ((1/30)*u.keV).astype(np.float32)
    l1.energies['e_high'][1:-1] = ((np.arange(31) / 30)[1:] * u.keV).astype(np.float32)
    l1.energies['e_low'][1:-1] = ((np.arange(31) / 30)[:-1] * u.keV).astype(np.float32)
    l1.data['time'] = l1.time_range.start + np.arange(5)/5*u.s
    count, count_err, times, timedel, energies = l1.get_data(time_indices=[[0, 4]],
                                                             energy_indices=[[1, 30]])
    assert np.allclose(count, 1 * u.ct / (u.s * u.keV))
    assert np.allclose(count_err, 1 * u.ct / (u.s * u.keV))


def test_science_l0():
    type_, num_times, num_detectors, num_pixels, num_energies = RawPixelData, 5, 32, 12, 32
    res = ScienceData.from_fits(test.STIX_SCI_XRAY_RPD)
    assert isinstance(res, type_)
    assert len(res.times) == num_times
    assert np.array_equal(res.detectors.masks, np.ones((1, num_detectors)))
    assert np.array_equal(res.pixels.masks, np.ones((1, num_pixels)))
    assert np.array_equal(res.energy_masks.masks, np.ones((1, num_energies)))


def test_science_l1():
    type_, num_times, num_detectors, num_pixels, num_energies = CompressedPixelData, 5, 32, 12, 32
    res = ScienceData.from_fits(test.STIX_SCI_XRAY_CPD)
    assert isinstance(res, type_)
    assert len(res.times) == num_times
    assert np.array_equal(res.detectors.masks, np.ones((1, num_detectors)))
    assert np.array_equal(res.pixels.masks, np.ones(12)[None, ...])
    assert np.array_equal(res.energy_masks.masks, np.ones((1, num_energies)))


def test_science_l2():
    type_, num_times, num_detectors, num_pixels, num_energies = SummedCompressedPixelData, 5, 32, 4, 32
    res = ScienceData.from_fits(test.STIX_SCI_XRAY_SCPD)
    assert isinstance(res, type_)
    assert len(res.times) == num_times
    assert np.array_equal(res.detectors.masks, np.ones((1, num_detectors)))
    # assert np.array_equal(res.pixels.masks.astype(int), np.ones((1, num_pixels)))
    assert np.array_equal(res.energy_masks.masks, np.ones((1, num_energies)))


# def test_science_l3():
#     type_, num_times, num_detectors, num_pixels, num_energies = Visibility, 5, 32, 12, 32
#     res = ScienceData.from_fits(test.STIX_SCI_XRAY_VIS)
#     assert isinstance(res, type_)
#     assert len(res.times) == num_times
#     assert np.array_equal(res.detectors.masks, np.ones((1, num_detectors)))
#     # assert np.array_equal(res.pixels.masks, np.ones((1, num_pixels)))
#     assert np.array_equal(res.energy_masks.masks, np.ones((1, num_energies)))


def test_spectrogram():
    type_, num_times, num_detectors, num_pixels, num_energies = Spectrogram, 5, 32, 12, 32
    res = ScienceData.from_fits(test.STIX_SCI_XRAY_SPEC)
    assert isinstance(res, type_)
    assert len(res.times) == num_times
    assert np.array_equal(res.detectors.masks, np.ones((1, num_detectors)))
    # assert np.array_equal(res.pixels.masks, np.ones((1, num_pixels)))
    assert np.array_equal(res.energy_masks.masks, np.ones((1, num_energies)))



import astropy.units as u
import numpy as np

from astropy.time import Time, TimeDelta
from numpy.testing import assert_equal


def times_to_indices(in_times, obs_times, unit=u.ms, decimals=3):
    """
    Convert the input time/s to indices.

    Parameters
    ----------
    in_times : `astropy.units.Quantity`
        A target duration e.g `5*u.s` an array of `Time` if 2d
    obs_times : `astropy.units.Quantity`

    unit : `astropy.units.Unit`

    decimals : int

    Returns
    -------
    Array of indices
    """
    unique_only = False
    shape = None
    relative_times = np.around((obs_times - obs_times[0]).to(unit), decimals=decimals)

    if in_times.isscalar:
        interval = in_times
        total_duration = obs_times[-1] - obs_times[0]
        if interval < total_duration:
            num_intervals = np.ceil((total_duration / interval).decompose()) + 1
            target_times = np.around((np.arange(num_intervals) * interval).to(unit),
                                     decimals=decimals)
            unique_only = True
        else:
            target_times = relative_times[[0,-1]]
    elif isinstance(in_times, Time):
        ndim = in_times.ndim
        target_times = np.around((in_times - obs_times[0]).to(unit), decimals=decimals)
        if ndim == 1:
            unique_only = True
        elif ndim == 2:
            shape = in_times.shape
            if 2 not in shape:
                raise ValueError('If 2d list given must have shape (2, x) or (x, 2).')
            if shape[0] == 2:
                target_times = target_times.T
                shape = shape[::-1]

            target_times = target_times.flatten()
        else:
            raise ValueError('Only 1d or 2d time arrays are supported')

    indices = np.abs(target_times[:, None] - relative_times[None, :]).argmin(axis=-1)
    i1 = _closest_index_loop(target_times, relative_times)
    i2 = _closest_index_broadcast(target_times, relative_times)
    i3 = _closest_index_searchsorted(target_times, relative_times)
    assert_equal(i1, i2)
    assert_equal(i2, i3)
    if shape:
        indices = indices.reshape(shape)
    if unique_only:
        indices = np.unique(indices)
    return indices


def _closest_index_loop(a, b):
    out = np.zeros(a.shape, int)
    for i, val in enumerate(a):
        out[i] = (np.abs(val-b)).argmin()
    return out

def _closest_index_searchsorted(a, b):
    L = b.size
    sidx_b = b.argsort()
    sorted_b = b[sidx_b]
    sorted_idx = np.searchsorted(sorted_b, a)
    sorted_idx[sorted_idx == L] = L - 1
    mask = (sorted_idx > 0) & ((np.abs(a - sorted_b[sorted_idx - 1]) <= np.abs(a - sorted_b[sorted_idx])))
    return sidx_b[sorted_idx - mask]

def _closest_index_broadcast(a, b):
    return np.abs(a[:, None] - b[None, :]).argmin(axis=-1)



from stixpy.product.product import L1Product


class HouseKeepingProduct(L1Product):
    """
    Basic HouseKeeping
    """


class HKMini(HouseKeepingProduct):
    """
    HK “mini” report only generated during the Start-up SW.
    """
    @classmethod
    def is_datasource_for(cls, *, meta, **kwargs):
        """Determines if meta data meach Raw Pixel Data"""
        service_subservice_ssid = tuple(meta[name] for name in ['STYPE', 'SSTYPE', 'SSID'])
        level = meta['level']
        if service_subservice_ssid == (3, 25, 1) and level == 'L1':
            return True

    def __repr__(self):
        return f'{self.__class__.__name__}\n' \
               f'    {self.time_range}'


class HKMaxi(HouseKeepingProduct):
    """
    HK “maxi” report with is reported in all modes of the Application SW. In Maintenance mode,
    however, all subsystems including Aspect System have to be disabled upon entry to the mode,
    so most of the HK will default to zero.
    """
    @classmethod
    def is_datasource_for(cls, *, meta, **kwargs):
        """Determines if meta data meach Raw Pixel Data"""
        service_subservice_ssid = tuple(meta[name] for name in ['STYPE', 'SSTYPE', 'SSID'])
        level = meta['level']
        if service_subservice_ssid == (3, 25, 2) and level == 'L1':
            return True

    def __repr__(self):
        return f'{self.__class__.__name__}\n' \
               f'    {self.time_range}'



from sunpy.time import TimeRange

from stixpy.product.product import L1Product


class QuickLookProduct(L1Product):
    """
    Basic science data class
    """
    @property
    def time_range(self):
        """
        A `sunpy.time.TimeRange` for the data.
        """
        return TimeRange(self.data['time'][0] - self.data['timedel'][0]/2,
                         self.data['time'][-1] + self.data['timedel'][-1]/2)


class QLLightCurve(QuickLookProduct):
    """
    Quicklook Lightcurves nominal in 5 energy bins every 4s
    """
    @classmethod
    def is_datasource_for(cls, *, meta, **kwargs):
        """Determines if meta data meach Raw Pixel Data"""
        service_subservice_ssid = tuple(meta[name] for name in ['STYPE', 'SSTYPE', 'SSID'])
        level = meta['level']
        if service_subservice_ssid == (21, 6, 30) and level == 'L1':
            return True

    def __repr__(self):
        return f'{self.__class__.__name__}\n' \
               f'    {self.time_range}'


class QLBackground(QuickLookProduct):
    """
    Quicklook Background nominal in 5 energy bins every 8s
    """
    @classmethod
    def is_datasource_for(cls, *, meta, **kwargs):
        """Determines if meta data meach Raw Pixel Data"""
        service_subservice_ssid = tuple(meta[name] for name in ['STYPE', 'SSTYPE', 'SSID'])
        level = meta['level']
        if service_subservice_ssid == (21, 6, 31) and level == 'L1':
            return True

    def __repr__(self):
        return f'{self.__class__.__name__}\n' \
               f'    {self.time_range}'



from collections import OrderedDict

import astropy.units as u
import matplotlib.dates
import numpy as np

from astropy.table import Table, QTable
from astropy.io import fits
from astropy.time.core import Time, TimeDelta
from astropy.visualization import quantity_support

from sunpy.visualization import peek_show
from sunpy.timeseries.timeseriesbase import GenericTimeSeries

__all__ = ['QLLightCurve', 'QLBackground', 'QLVariance', 'HKMaxi']


eta = 2.5 * u.us
tau = 12.5 * u.us


def _lfrac(trigger_rate):
    nin = trigger_rate / (1 - (trigger_rate * (tau+eta)))
    return np.exp(-eta * nin) / (1 + tau * nin)


class QLLightCurve(GenericTimeSeries):
    """
    Quicklook X-ray time series.

    Nominally in 5 energy bands from 4 - 150 kev

    Examples
    --------
    >>> from stixpy.data import test
    >>> from sunpy.timeseries import TimeSeries
    >>> from stixpy.timeseries.quicklook import QLLightCurve
    >>> ql_lc = TimeSeries(test.STIX_QL_LIGHTCURVE_TIMESERIES)
    >>> ql_lc # doctest: +SKIP
    QLLightCurve
           <sunpy.time.timerange.TimeRange object at ...>
        Start: 2020-05-06 00:00:01
        End:   2020-05-06 23:59:57
        Center:2020-05-06 11:59:59
        Duration:0.9999538194444444 days or
               23.998891666666665 hours or
               1439.9334999999999 minutes or
               86396.01 seconds
    <BLANKLINE>
    """
    _source = 'stix'

    @peek_show
    def peek(self, **kwargs):
        """
        Displays a graphical overview of the data in this object for user evaluation.
        For the creation of plots, users should instead use the
        `.plot` method and Matplotlib's pyplot framework.

        Parameters
        ----------
        **kwargs : `dict`
            Any additional plot arguments that should be used when plotting.
        """
        import matplotlib.pyplot as plt
        # Check we have a timeseries valid for plotting
        self._validate_data_for_plotting()
        self.plot(**kwargs)

    def plot(self, axes=None, **plot_args):
        """
        Show a plot of the data.

        Parameters
        ----------
        axes : `~matplotlib.axes.Axes`, optional
            If provided the image will be plotted on the given axes.
            Defaults to `None`, so the current axes will be used.
        **plot_args : `dict`, optional
            Additional plot keyword arguments that are handed to
            :meth:`pandas.DataFrame.plot`.

        Returns
        -------
        axes : `~matplotlib.axes.Axes`
            The plot axes.
        """
        import matplotlib.pyplot as plt
        # Get current axes
        if axes is None:
            fig, axes = plt.subplots()
        else:
            fig = plt.gcf()

        self._validate_data_for_plotting()
        quantity_support()

        dates = matplotlib.dates.date2num(self.to_dataframe().index)

        labels = [f'{col}' for col in self.columns[5:]]

        lines = [axes.plot_date(dates, self.to_dataframe().iloc[:, 5+i], '-', label=labels[i], **plot_args)
         for i in range(5)]

        axes.legend(loc='upper right')

        axes.set_yscale("log")

        axes.set_title('STIX Quick Look')
        axes.set_ylabel('count s$^{-1}$ keV$^{-1}$')

        axes.yaxis.grid(True, 'major')
        axes.xaxis.grid(False, 'major')
        axes.legend()

        # TODO: display better tick labels for date range (e.g. 06/01 - 06/05)
        formatter = matplotlib.dates.DateFormatter('%d %H:%M')
        axes.xaxis.set_major_formatter(formatter)

        axes.fmt_xdata = matplotlib.dates.DateFormatter('%d %H:%M')
        fig.autofmt_xdate()

        return lines

    @classmethod
    def _parse_file(cls, filepath):
        """
        Parses a GOES/XRS FITS file.

        Parameters
        ----------
        filepath : `str`
            The path to the file you want to parse.
        """
        hdus = fits.open(filepath)
        return cls._parse_hdus(hdus)

    @classmethod
    def _parse_hdus(cls, hdulist):
        """
        Parses STIX FITS data files to create TimeSeries.

        Parameters
        ----------
        hdulist :
        """
        header = hdulist[0].header
        control = QTable(hdulist[1].data)
        header['control'] = control
        data = QTable(hdulist[2].data)
        energies = QTable(hdulist[4].data)
        energy_delta = energies['e_high'] - energies['e_low'] << u.keV

        live_time = _lfrac(data['triggers'].reshape(-1)/(16*data['timedel']*u.s))

        data['counts'] = data['counts'] / (live_time.reshape(-1, 1) * energy_delta)

        names = ['{:d}-{:d} keV'.format(energies["e_low"][i].astype(int),energies["e_high"][i].astype(int)) for i in range(5)]

        [data.add_column(data['counts'][:, i], name=names[i]) for i in range(5)]
        data.remove_column('counts')
        [data.add_column(data['counts_comp_err'][:, i], name=f'{names[i]}_comp_err') for i in range(5)]
        data.remove_column('counts_comp_err')

        try:
            data['time'] = Time(header['date_obs']) + data['time'] * u.cs
        except KeyError:
            data['time'] = Time(header['date-obs']) + TimeDelta(data['time'] * u.cs)

        units = OrderedDict([('control_index', None),
                             ('timedel', u.s),
                             ('triggers', None),
                             ('triggers_comp_err', None),
                             ('rcr', None)])
        units.update([(name, u.ct) for name in names])
        units.update([(f'{name}_comp_err', u.ct) for name in names])

        data['triggers'] = data['triggers'].reshape(-1)
        data['triggers_comp_err'] = data['triggers_comp_err'].reshape(-1)

        data_df = data.to_pandas()
        data_df.index = data_df['time']
        data_df.drop(columns='time', inplace=True)

        return data_df, header, units

    @classmethod
    def is_datasource_for(cls, **kwargs):
        """
        Determines if the file corresponds to a STIX QL LightCurve
        `~sunpy.timeseries.TimeSeries`.
        """
        # Check if source is explicitly assigned
        if 'source' in kwargs.keys():
            if kwargs.get('source', ''):
                return kwargs.get('source', '').lower().startswith(cls._source)
        # Check if HDU defines the source instrument
        if 'meta' in kwargs.keys():
            return (kwargs['meta'].get('telescop', '') == 'SOLO/STIX'
                    and 'ql-lightcurve' in kwargs['meta']['filename'])

    def __repr__(self):
        return f'{self.__class__.__name__}\n' \
               f'    {self.time_range}'


class QLBackground(GenericTimeSeries):
    """
    Quicklook X-ray background detector time series.

    Examples
    --------
    >>> from stixpy.data import test
    >>> from sunpy.timeseries import TimeSeries
    >>> from stixpy.timeseries.quicklook import QLLightCurve
    >>> ql_bg = TimeSeries(test.STIX_QL_BACKGROUND_TIMESERIES)
    >>> ql_bg # doctest: +SKIP
    <stixpy.timeseries.quicklook.QLBackground object at ...
    SunPy TimeSeries
    ----------------
     Observatory:		 SOLO/STIX
     Instrument:		 STIX
     Channel(s):		 control_index<br>timedel<br>triggers<br>triggers_err<br>4.0-10.0<br>10.0-15.0<br>15.0-25.0<br>25.0-50.0<br>50.0-150.0<br>4.0-10.0_err<br>10.0-15.0_err<br>15.0-25.0_err<br>25.0-50.0_err<br>50.0-150.0_err
     Start Date:		 2020-05-06 00:00:04
     End Date:		 2020-05-06 23:59:56
     Center Date:		 2020-05-06 11:59:59
     Resolution:		 8.03 s
     Samples per Channel:		 10758
     Data Range(s):		 control_index    1.60E+01<br>timedel          0.00E+00<br>triggers         1.10E+03<br>triggers_err     3.23E+01<br>4.0-10.0         7.72E+01<br>10.0-15.0        1.34E+02<br>15.0-25.0        6.07E+01<br>25.0-50.0        2.17E+01<br>50.0-150.0       4.95E+00<br>4.0-10.0_err     9.25E+00<br>10.0-15.0_err    1.85E+01<br>15.0-25.0_err    1.85E+01<br>25.0-50.0_err    1.85E+01<br>50.0-150.0_err   9.25E+00
     Units:		 None<br>ct<br>s
                              control_index  timedel  ...  25.0-50.0_err  50.0-150.0_err
     time                                             ...
     2020-05-06 00:00:03.531              3      800  ...       2.345208        0.707107
     2020-05-06 00:00:11.531              3      800  ...       2.345208        0.707107
     2020-05-06 00:00:19.531              3      800  ...       2.345208        0.000000
     2020-05-06 00:00:27.531              3      800  ...       2.345208        0.000000
     2020-05-06 00:00:35.531              3      800  ...       2.345208        0.000000
     ...                                ...      ...  ...            ...             ...
     2020-05-06 23:59:23.531              2      800  ...       0.707107        1.224745
     2020-05-06 23:59:31.531              2      800  ...       0.000000        2.345208
     2020-05-06 23:59:39.531              2      800  ...       0.000000        4.636809
     2020-05-06 23:59:47.531              2      800  ...       0.000000        1.224745
     2020-05-06 23:59:55.531              2      800  ...       0.707107        2.345208
     <BLANKLINE>
     [10758 rows x 14 columns]

    """
    @peek_show
    def peek(self, **kwargs):
        """
        Displays a graphical overview of the data in this object for user evaluation.
        For the creation of plots, users should instead use the
        `.plot` method and Matplotlib's pyplot framework.

        Parameters
        ----------
        **kwargs : `dict`
            Any additional plot arguments that should be used when plotting.
        """
        import matplotlib.pyplot as plt
        # Check we have a timeseries valid for plotting
        self._validate_data_for_plotting()
        self.plot(**kwargs)

    def plot(self, axes=None, **plot_args):
        """
        Show a plot of the data.

        Parameters
        ----------
        axes : `~matplotlib.axes.Axes`, optional
            If provided the image will be plotted on the given axes.
            Defaults to `None`, so the current axes will be used.
        **plot_args : `dict`, optional
            Additional plot keyword arguments that are handed to
            :meth:`pandas.DataFrame.plot`.

        Returns
        -------
        axes : `~matplotlib.axes.Axes`
            The plot axes.
        """
        import matplotlib.pyplot as plt
        # Get current axes
        if axes is None:
            fig, axes = plt.subplots()

        self._validate_data_for_plotting()
        quantity_support()

        dates = matplotlib.dates.date2num(self.to_dataframe().index)

        labels = [f'{col} keV' for col in self.columns[4:]]

        [axes.plot_date(dates, self.to_dataframe().iloc[:, 4+i], '-', label=labels[i], **plot_args)
         for i in range(5)]

        axes.legend(loc='upper right')

        axes.set_yscale("log")

        axes.set_title('STIX Quick Look')
        axes.set_ylabel('count s$^{-1}$ keV$^{-1}$')

        axes.yaxis.grid(True, 'major')
        axes.xaxis.grid(False, 'major')
        axes.legend()

        # TODO: display better tick labels for date range (e.g. 06/01 - 06/05)
        formatter = matplotlib.dates.DateFormatter('%d %H:%M')
        axes.xaxis.set_major_formatter(formatter)

        axes.fmt_xdata = matplotlib.dates.DateFormatter('%d %H:%M')
        fig.autofmt_xdate()

        return fig

    @classmethod
    def _parse_file(cls, filepath):
        """
        Parses a GOES/XRS FITS file.

        Parameters
        ----------
        filepath : `str`
            The path to the file you want to parse.
        """
        hdus = fits.open(filepath)
        return cls._parse_hdus(hdus)

    @classmethod
    def _parse_hdus(cls, hdulist):
        """
        Parses STIX FITS data files to create TimeSeries.

        Parameters
        ----------
        hdulist :
            The path to the file you want to parse.
        """
        header = hdulist[0].header
        control = Table(hdulist[1].data)
        header['control'] = control
        data = Table(hdulist[2].data)
        energies = Table(hdulist[4].data)

        energy_delta = energies['e_high'] - energies['e_low'] << u.keV

        live_time = _lfrac(data['triggers'].reshape(-1)/(16*data['timedel']*u.s))

        data['counts'] = data['counts'] / (live_time.reshape(-1, 1) * energy_delta)

        names = [f'{energies["e_low"][i]}-{energies["e_high"][i]}' for i in range(5)]

        [data.add_column(data['counts'][:, i], name=names[i]) for i in range(5)]
        data.remove_column('counts')

        [data.add_column(data['counts_comp_err'][:, i], name=f'{names[i]}_comp_err') for i in range(5)]
        data.remove_column('counts_comp_err')

        try:
            data['time'] = Time(header['date_obs']) + TimeDelta(data['time'] * u.s)
        except KeyError:
            data['time'] = Time(header['date-obs']) + TimeDelta(data['time'] * u.s)

        # [f'{energies[i]["e_low"]} - {energies[i]["e_high"]} keV' for i in range(5)]

        units = OrderedDict([('control_index', None),
                             ('timedel', u.s),
                             ('triggers', None),
                             ('triggers_comp_err', None),
                             ('rcr', None)])
        units.update([(name, u.ct) for name in names])
        units.update([(f'{name}_comp_err', u.ct) for name in names])

        data['triggers'] = data['triggers'].reshape(-1)
        data['triggers_comp_err'] = data['triggers_comp_err'].reshape(-1)

        data_df = data.to_pandas()
        data_df.index = data_df['time']
        data_df.drop(columns='time', inplace=True)

        return data_df, header, units

    @classmethod
    def is_datasource_for(cls, **kwargs):
        """
        Determines if the file corresponds to a STIX QL LightCurve
        `~sunpy.timeseries.TimeSeries`.
        """
        # Check if source is explicitly assigned
        if 'source' in kwargs.keys():
            if kwargs.get('source', ''):
                return kwargs.get('source', '').lower().startswith(cls._source)
        # Check if HDU defines the source instrument
        if 'meta' in kwargs.keys():
            return (kwargs['meta'].get('telescop', '') == 'SOLO/STIX'
                    and 'ql-background' in kwargs['meta']['filename'])


class QLVariance(GenericTimeSeries):
    """
    Quicklook X-ray background detector time series.

    Examples
    --------
    >>> from stixpy.data import test
    >>> from sunpy.timeseries import TimeSeries
    >>> from stixpy.timeseries.quicklook import QLLightCurve
    >>> ql_var = TimeSeries(test.STIX_QL_VARIANCE_TIMESERIES)
    >>> ql_var # doctest: +SKIP
    <stixpy.timeseries.quicklook.QLVariance object at ...>
    SunPy TimeSeries
    ----------------
    Observatory:		 SOLO/STIX
    Instrument:		     STIX
    Channel(s):		     timedel<br>control_index<br>4.0-20.0<br>4.0-20.0_err
    Start Date:		     2020-05-06 00:00:02
    End Date:		  2020-05-06 23:59:58
    Center Date:		 2020-05-06 11:59:59
    Resolution:		 4.015 s
    Samples per Channel:		 21516
    Data Range(s):		 timedel         0.00E+00<br>control_index   6.00E+00<br>4.0-20.0        4.93E+00<br>4.0-20.0_err    5.87E+02
    Units:		 ct<br>s<br>None
                             timedel  control_index  4.0-20.0  4.0-20.0_err
    time
    2020-05-06 00:00:01.531      400              2  0.419844     73.901962
    2020-05-06 00:00:05.531      400              2  0.169844     36.952671
    2020-05-06 00:00:09.531      400              2  0.154844     18.479719
    2020-05-06 00:00:13.531      400              2  1.359844    295.603607
    2020-05-06 00:00:17.531      400              2  0.114844     18.479719
    ...                          ...            ...       ...           ...
    2020-05-06 23:59:41.531      400              1  0.919844    147.802231
    2020-05-06 23:59:45.531      400              1  0.919844    147.802231
    2020-05-06 23:59:49.531      400              1  2.159844    295.603607
    2020-05-06 23:59:53.531      400              1  0.094844     18.479719
    2020-05-06 23:59:57.531      400              1  0.154844     18.479719
    <BLANKLINE>
    [21516 rows x 4 columns]
    """
    def plot(self, axes=None, **plot_args):
        """
        Show a plot of the data.

        Parameters
        ----------
        axes : `~matplotlib.axes.Axes`, optional
            If provided the image will be plotted on the given axes.
            Defaults to `None`, so the current axes will be used.
        **plot_args : `dict`, optional
            Additional plot keyword arguments that are handed to
            :meth:`pandas.DataFrame.plot`.

        Returns
        -------
        axes : `~matplotlib.axes.Axes`
            The plot axes.
        """
        import matplotlib.pyplot as plt
        # Get current axes
        if axes is None:
            fig, axes = plt.subplots()

        self._validate_data_for_plotting()
        quantity_support()

        dates = matplotlib.dates.date2num(self.to_dataframe().index)

        label = f'{self.columns[2]} keV'

        axes.plot_date(dates, self.to_dataframe().iloc[:, 2], '-', label=label) #, **plot_args)

        axes.legend(loc='upper right')

        axes.set_yscale("log")

        axes.set_title('STIX Quick Look Variance')
        axes.set_ylabel('count s$^{-1}$ keV$^{-1}$')

        axes.yaxis.grid(True, 'major')
        axes.xaxis.grid(False, 'major')
        axes.legend()

        # TODO: display better tick labels for date range (e.g. 06/01 - 06/05)
        formatter = matplotlib.dates.DateFormatter('%d %H:%M')
        axes.xaxis.set_major_formatter(formatter)

        axes.fmt_xdata = matplotlib.dates.DateFormatter('%d %H:%M')
        fig.autofmt_xdate()

        return fig

    @classmethod
    def _parse_file(cls, filepath):
        """
        Parses a GOES/XRS FITS file.

        Parameters
        ----------
        filepath : `str`
            The path to the file you want to parse.
        """
        hdus = fits.open(filepath)
        return cls._parse_hdus(hdus)

    @classmethod
    def _parse_hdus(cls, hdulist):
        """
        Parses STIX FITS data files to create TimeSeries.

        Parameters
        ----------
        hdulist :
            The path to the file you want to parse.
        """
        header = hdulist[0].header
        control = Table(hdulist[1].data)
        header['control'] = control
        data = Table(hdulist[2].data)
        energies = Table(hdulist[4].data)
        dE = energies[control['energy_bin_mask'][0]]['e_high'][-1] \
             - energies[control['energy_bin_mask'][0]]['e_low'][0] << u.keV
        name = f'{energies[control["energy_bin_mask"][0]]["e_low"][0]}' \
               f'-{energies[control["energy_bin_mask"][0]]["e_high"][-1]}'

        try:
            data['time'] = Time(header['date_obs']) + TimeDelta(data['time'] * u.s)
        except KeyError:
            data['time'] = Time(header['date-obs']) + TimeDelta(data['time'] * u.s)

        data['variance'] = data['variance'].reshape(-1)/(dE * data['timedel'])

        data.add_column(data['variance'], name=name)
        data.remove_column('variance')
        data.add_column(data['variance_comp_err'], name=f'{name}_comp_err')
        data.remove_column('variance_comp_err')

        units = OrderedDict([('control_index', None),
                             ('timedel', u.s),
                             (name, u.count),
                             (f'{name}_comp_err', u.count)])

        data[name] = data[name].reshape(-1)
        data[f'{name}_comp_err'] = data[f'{name}_comp_err'].reshape(-1)

        data_df = data.to_pandas()
        data_df.index = data_df['time']
        data_df.drop(columns='time', inplace=True)

        return data_df, header, units

    @classmethod
    def is_datasource_for(cls, **kwargs):
        """
        Determines if the file corresponds to a STIX QL LightCurve
        `~sunpy.timeseries.TimeSeries`.
        """
        # Check if source is explicitly assigned
        if 'source' in kwargs.keys():
            if kwargs.get('source', ''):
                return kwargs.get('source', '').lower().startswith(cls._source)
        # Check if HDU defines the source instrument
        if 'meta' in kwargs.keys():
            return (kwargs['meta'].get('telescop', '') == 'SOLO/STIX'
                    and 'ql-variance' in kwargs['meta']['filename'])


class HKMaxi(GenericTimeSeries):
    """
    Housekeeping Maxi Report.

    Examples
    --------
    >>> from stixpy.data import test
    >>> from sunpy.timeseries import TimeSeries
    >>> from stixpy.timeseries.quicklook import HKMaxi
    >>> hk = TimeSeries(test.STIX_HK_MAXI_TIMESERIES)
    >>> hk  # doctest: +SKIP
    <stixpy.timeseries.quicklook.HKMaxi ...>
    """
    # def plot(self, axes=None, **plot_args):
    #     """
    #     Show a plot of the data.
    #
    #     Parameters
    #     ----------
    #     axes : `~matplotlib.axes.Axes`, optional
    #         If provided the image will be plotted on the given axes.
    #         Defaults to `None`, so the current axes will be used.
    #     **plot_args : `dict`, optional
    #         Additional plot keyword arguments that are handed to
    #         :meth:`pandas.DataFrame.plot`.
    #
    #     Returns
    #     -------
    #     axes : `~matplotlib.axes.Axes`
    #         The plot axes.
    #     """
    #     import matplotlib.pyplot as plt
    #     # Get current axes
    #     if axes is None:
    #         fig, axes = plt.subplots()
    #
    #     self._validate_data_for_plotting()
    #     quantity_support()
    #
    #     dates = matplotlib.dates.date2num(self.to_dataframe().index)
    #
    #     label = f'{self.columns[2]} keV'
    #
    #     axes.plot_date(dates, self.to_dataframe().iloc[:, 2], '-', label=label) #, **plot_args)
    #
    #     axes.legend(loc='upper right')
    #
    #     axes.set_yscale("log")
    #
    #     axes.set_title('STIX Quick Look Variance')
    #     axes.set_ylabel('count s$^{-1}$ keV$^{-1}$')
    #
    #     axes.yaxis.grid(True, 'major')
    #     axes.xaxis.grid(False, 'major')
    #     axes.legend()
    #
    #     # TODO: display better tick labels for date range (e.g. 06/01 - 06/05)
    #     formatter = matplotlib.dates.DateFormatter('%d %H:%M')
    #     axes.xaxis.set_major_formatter(formatter)
    #
    #     axes.fmt_xdata = matplotlib.dates.DateFormatter('%d %H:%M')
    #     fig.autofmt_xdate()
    #
    #     return fig

    @classmethod
    def _parse_file(cls, filepath):
        """
        Parses a GOES/XRS FITS file.

        Parameters
        ----------
        filepath : `str`
            The path to the file you want to parse.
        """
        hdus = fits.open(filepath)
        return cls._parse_hdus(hdus)

    @classmethod
    def _parse_hdus(cls, hdulist):
        """
        Parses STIX FITS data files to create TimeSeries.

        Parameters
        ----------
        hdulist :
            The path to the file you want to parse.
        """
        header = hdulist[0].header
        control = Table(hdulist[1].data)
        header['control'] = control
        data = Table(hdulist[2].data)

        try:
            data['time'] = Time(header['date_obs']) + TimeDelta(data['time'] * u.s)
        except KeyError:
            data['time'] = Time(header['date-obs']) + TimeDelta(data['time'] * u.s)

        data_df = data.to_pandas()
        data_df.index = data_df['time']
        data_df.drop(columns='time', inplace=True)

        return data_df, header, None

    @classmethod
    def is_datasource_for(cls, **kwargs):
        """
        Determines if the file corresponds to a STIX QL LightCurve
        `~sunpy.timeseries.TimeSeries`.
        """
        # Check if source is explicitly assigned
        if 'source' in kwargs.keys():
            if kwargs.get('source', ''):
                return kwargs.get('source', '').lower().startswith(cls._source)
        # Check if HDU defines the source instrument
        if 'meta' in kwargs.keys():
            return (kwargs['meta'].get('telescop', '') == 'SOLO/STIX'
                    and 'hk-maxi' in kwargs['meta']['filename'])



"""

Examples
-------
An AIA map

.. plot::
   :include-source: true

   import astropy.units as u
   from astropy.coordinates import SkyCoord
   from sunpy.net import Fido, attrs as a
   from sunpy.map import Map
   from sunpy.coordinates.frames import HeliographicStonyhurst
   from stixpy.visualisation.map_reprojection import reproject_map, plot_map_reproj, get_solo_position

   # Search and download map using FIDO
   aia = (a.Instrument.aia &
       a.Sample(24 * u.hour) &
       a.Time('2021-04-17', '2021-04-18'))
   wave = a.Wavelength(19.3 * u.nm, 19.3 * u.nm)
   query = Fido.search(wave, aia)
   results = Fido.fetch(query[0][0])

   # Create map and resample to speed up calculations
   map = Map(results)
   map = map.resample([512, 512]*u.pix)

   # Get SOLO position observer
   # observer = get_solo_position(map) # broken on RTD
   observer = SkyCoord(-97.01771373*u.deg, 21.20931737*u.deg, 1.25689226e+08*u.km,
                       frame=HeliographicStonyhurst, obstime='2021-04-17 00:00:04.840000')

   # Reproject Map
   reprojected_map = reproject_map(map, observer, out_shape=(768, 768))

   # Run plotting function
   plot_map_reproj(map, reprojected_map)


A HMI map

.. plot::
   :include-source: true

   import astropy.units as u
   from astropy.coordinates import SkyCoord
   from sunpy.net import Fido, attrs as a
   from sunpy.map import Map
   from sunpy.coordinates.frames import HeliographicStonyhurst
   from stixpy.visualisation.map_reprojection import reproject_map, plot_map_reproj, get_solo_position

   # Search and download map using FIDO
   query = Fido.search(a.Time('2020/06/12 13:20:00', '2020/06/12 13:40:00'),
                       a.Instrument.hmi, a.Physobs.los_magnetic_field)
   result = Fido.fetch(query[0][0])

   # Create map and resample to speed up calculations
   map = Map(result)
   map = map.resample([512, 512] * u.pix)

   # Set SOLO as observer
   # observer = get_solo_position(map) # broke on RTD
   observer = SkyCoord(56.18727061*u.deg, 3.5358653*u.deg, 77369481.8542484*u.km,
                       frame=HeliographicStonyhurst, obstime='2020-06-12 13:20:08.300000')

   # Reproject Map
   reprojected_map = reproject_map(map, observer, out_shape=(1024, 1024))

   # Run plotting function
   plot_map_reproj(map, reprojected_map)


"""

import astropy.units as u
import matplotlib.pyplot as plt

import sunpy
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from reproject import reproject_interp
try:
    from stixcore.ephemeris.manager import Spice
except (ImportError, ModuleNotFoundError):
    Spice = None
from sunpy.coordinates import frames
from sunpy.coordinates import get_body_heliographic_stonyhurst

from stixpy.utils.logging import get_logger

logger = get_logger(__name__)

__all__ = ['get_solo_position', 'reproject_map', 'create_headers', 'plot_map_reproj']


def get_solo_position(map):
    """
    Return the position of SOLO at the time the map was observed

    Parameters
    ----------
    map : `sunpy.map.Map`
        Map to reproject to be as seen from SOLO

    Returns
    -------
    solo_hgs : `astropy.coordinates.SkyCoord`
        The position of SOLO in HeliographicStonyhurst frame
    """
    if Spice is not None:
        logger.info('Using Spice')
        p = Spice.instance.get_position(date=map.date.datetime, frame='SOLO_HEE')
        solo_hee = SkyCoord(*p, frame=frames.HeliocentricEarthEcliptic, representation_type='cartesian',
                            obstime=map.date.datetime)
        # Converting HeliocentricEarthEcliptic coords of SOLAR ORBITER position to
        # HeliographicStonyhurst frame
        solo_hgs = solo_hee.transform_to(frames.HeliographicStonyhurst)
    else:
        logger.info('Spice not configured falling back to JPL ')
        solo_hgs = get_horizons_coord('solo', time=map.date)


def create_headers(obs_ref_coord, map, out_shape=None, out_scale=None):
    """
    Generates MetaDict and WCS headers for reprojected map

    Parameters
    ----------
    obs_ref_coord : `sunpy.map.Map`
        Target WCS reference coordinate (as seen by observer).
        Generated in map_reproject_to_observer function.

    out_shape : `tuple`
        The shape of the reprojected map - defaults to (512, 512)

    out_scale : `Quantity`
        The scale of the output map

    Returns
    -------
    obs_wcs_header : `astropy.wcs`
        WCS header for reprojected map
    obs_metadict_header : `MetaDict`
        MetaDict header for reprojected map
    """
    if out_scale is None:
        out_scale = u.Quantity(map.scale)

    if out_shape is None:
        out_shape = map.data.shape

    if map.wavelength.unit.to_string() == '':
        obs_metadict_header = sunpy.map.make_fitswcs_header(
            out_shape,
            obs_ref_coord,
            scale=out_scale,
            rotation_matrix=map.rotation_matrix,
            instrument=map.detector)
    else:
        obs_metadict_header = sunpy.map.make_fitswcs_header(
            out_shape,
            obs_ref_coord,
            scale=out_scale,
            rotation_matrix=map.rotation_matrix,
            instrument=map.detector,
            wavelength=map.wavelength)
    obs_wcs_header = WCS(obs_metadict_header)
    return obs_wcs_header, obs_metadict_header


def reproject_map(map, observer, out_shape=None):
    """
    Reproject a map as viewed from a different observer.

    Parameters
    ----------
    map : `sunpy.map.Map`
        The input map to be reprojected
    observer : `astropy.coordinates.SkyCoord`
        The coordinates of the observer in HeliographicStonyhurst frame
    out_shape : `tuple`
        The shape of the reprojected map - defaults to same size as input map

    Returns
    -------
    map : `sunpy.map.Map`
        Reprojected map
    """
    if out_shape is None:
        out_shape = map.data.shape

    obs_ref_coord = SkyCoord(map.reference_coordinate.Tx, map.reference_coordinate.Ty,
                             obstime=map.reference_coordinate.obstime,
                             observer=observer, frame="helioprojective")
    obs_wcs_header, obs_metadict_header = create_headers(obs_ref_coord, map, out_shape=out_shape)
    output, footprint = reproject_interp(map, obs_wcs_header, out_shape)
    outmap = sunpy.map.Map(output, obs_metadict_header)
    outmap.plot_settings = map.plot_settings
    return outmap


def plot_map_reproj(map, reprojected_map):
    """
    Plot the original map, reprojected map and observer locations

    Parameters
    ----------
    map : `sunpy.map.Map`
        The input map to be reprojected
    reprojected_map :  `sunpy.map.Map`
        The reprojected map

    Returns
    -------
    `matplotlib.Figure`
        Figure showing the original map, reprojected map, and the observer locations
    """
    fig = plt.figure(figsize=(16, 4))
    ax1 = fig.add_subplot(1, 3, 1, projection=map)
    map.plot(axes=ax1, title=f'Input {map.detector} map {map.date}')
    reprojected_map.draw_grid(annotate=False, color='w')
    ax2 = fig.add_subplot(1, 3, 2, projection=reprojected_map)
    reprojected_map.plot(axes=ax2, title=f'Map as seen by observer {map.date}')
    reprojected_map.draw_grid(annotate=False, color='k')
    ax2.axes.get_yaxis().set_visible(False)

    new_observer = reprojected_map.observer_coordinate
    original_observer = map.observer_coordinate
    # Plotting position of the Sun
    sun_coords = get_body_heliographic_stonyhurst("sun", map.date)

    # Plotting polar positions
    ax3 = fig.add_subplot(1, 3, 3, projection="polar")
    ax3.plot(new_observer.lon.to(u.rad),
             new_observer.radius.to(u.AU),
             'o', ms=10,
             label="New Map Observer")
    ax3.plot(original_observer.lon.to(u.rad),
             original_observer.radius.to(u.AU),
             'o', ms=10,
             label="Original Map Observer")
    ax3.plot(sun_coords.lon.to(u.rad),
             sun_coords.radius.to(u.AU),
             'o', ms=10,
             label="Sun")
    fig.legend()
    plt.show()
    return fig



from astropy.utils import deprecated

from stixpy.product import Product


class StixpyDepreciationWarning(Warning):
    """
    A warning class to indicate a deprecated feature.
    """


@deprecated(since='0.1.0', alternative='stixpy.product.Product',
            warning_type=StixpyDepreciationWarning)
class ScienceData:
    @deprecated(since='0.1.0', alternative='stixpy.product.Product',
                warning_type=StixpyDepreciationWarning)
    @classmethod
    def from_fits(cls, file):
        return Product(file)



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
from itertools import chain

from setuptools import setup
from setuptools.config import read_configuration

################################################################################
# Programmatically generate some extras combos.
################################################################################
extras = read_configuration("setup.cfg")['options']['extras_require']

# Dev is everything
extras['dev'] = list(chain(*extras.values()))

# All is everything but tests and docs
exclude_keys = ("tests", "docs", "dev")
ex_extras = dict(filter(lambda i: i[0] not in exclude_keys, extras.items()))
# Concatenate all the values together for 'all'
extras['all'] = list(chain.from_iterable(ex_extras.values()))

################################################################################
# Version configuration and setup call
################################################################################

VERSION_TEMPLATE = """
# Note that we need to fall back to the hard-coded version if either
# setuptools_scm can't be imported or setuptools_scm can't determine the
# version, so we catch the generic 'Exception'.
try:
    from setuptools_scm import get_version
    __version__ = get_version(root='..', relative_to=__file__)
except Exception:
    __version__ = '{version}'
""".lstrip()

setup(
    extras_require=extras,
    use_scm_version={'write_to': os.path.join('stixpy', 'version.py'),
                     'write_to_template': VERSION_TEMPLATE},

)



from astropy.time import Time


class BaseProduct:
    """
    Base Product that all other product inherit from contains the registry for the factory pattern
    """
    _registry = {}

    def __init_subclass__(cls, **kwargs):
        """__init_subclass__ hook initializes all of the subclasses of a given class.

        So for each subclass, it will call this block of code on import.
        This registers each subclass in a dict that has the `is_datasource_for` attribute.
        This is then passed into the factory so we can register them.
        """
        super().__init_subclass__(**kwargs)
        if hasattr(cls, 'is_datasource_for'):
            cls._registry[cls] = cls.is_datasource_for


class GenericProduct(BaseProduct):
    def __init__(self, *, meta, control, data, idb_versions=None, energies=None):
        """
        Generic product composed of meta, control, data and optionally idb, and energy information

        Parameters
        ----------
        control : stix_parser.products.quicklook.Control
            Table containing control information
        data : stix_parser.products.quicklook.Data
            Table containing data
        """
        self.meta = meta
        self.control = control
        self.data = data
        self.idb_versions = idb_versions
        self._energies = energies

    @property
    def level(self):
        return self.meta.get('level')

    @property
    def service_type(self):
        return self.meta.get('stype')

    @property
    def servie_subtype(self):
        return self.meta.get('stype')

    @property
    def ssid(self):
        return self.meta.get('ssid')


class LevelBinary(GenericProduct):
    """Level Binary data"""
    def __init__(self, **kwargs):
        raise NotImplementedError('Level Binary Data is not currently supported in stixpy')

    @classmethod
    def is_datasource_for(cls, *, meta, **kwargs):
        """Determines if meta data meach Raw Pixel Data"""
        level = meta['level']
        if level == 'LB':
            return True

class L1Product(GenericProduct):
    """Level Binary data"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        meta = kwargs['meta']
        data = kwargs['data']

        # TODO don't change the data add new property or similar
        try:
            data['time'] = Time(meta['date-obs']) + data['time']
        except KeyError:
            data['time'] = Time(meta['date_obs']) + data['time']

class Level2(GenericProduct):
    """Level Binary data"""

    def __init__(self, **kwargs):
        raise NotImplementedError('Level Binary Data is not currently supported in stixpy')

    @classmethod
    def is_datasource_for(cls, *, meta, **kwargs):
        """Determines if meta data meach Raw Pixel Data"""
        level = meta['level']
        if level == 'LB':
            return True



import pathlib
from urllib.request import Request

import numpy as np

import astropy.io.fits
from astropy.io import fits
from astropy.table import QTable

from sunpy.data import cache

from sunpy.util import expand_list
from sunpy.util.datatype_factory_base import (
    BasicRegistrationFactory,
    MultipleMatchError,
    NoMatchError,
    ValidationFunctionError,
)
from sunpy.util.exceptions import NoMapsInFileError, warn_user
from sunpy.util.functools import seconddispatch
from sunpy.util.io import is_url, parse_path, possibly_a_path
from sunpy.util.metadata import MetaDict

from stixpy.product.product import GenericProduct
from stixpy.utils.logging import get_logger
from stixpy.product.sources import *


__all__ = ['Product', 'ProductFactory']

BITS_TO_UINT = {
    8: np.ubyte,
    16: np.uint16,
    32: np.uint32,
    64: np.uint64
}


logger = get_logger(__name__)


def read_qtable(file, hdu, hdul=None):
    """
    Read a fits extension into a QTable and maintain dtypes of columns with units

    Hack to work around QTable not respecting the dtype in fits file see
    https://github.com/astropy/astropy/issues/12494

    Parameters
    ----------
    input : `str`, `pathlib.Path`, `astropy.io.fits.HDUList` or `astropy.io.fits.HDU`
        Fits file
    hdu : int or str, optional
        The HDU to read the table from.
    hdul : `astropy.io.fits.HDUList`

    Returns
    -------
    `astropy.table.QTable`
        The corrected QTable with correct data types
    """
    if hdul is None:
        hdul = fits.open(file)

    qtable = QTable.read(file, hdu)

    for col in hdul[hdu].data.columns:
        if col.unit:
            logger.debug(f'Unit present dtype correction needed for {col}')
            dtype = col.dtype

            if col.bzero:
                logger.debug(f'Unit present dtype and bzero correction needed for {col}')
                bits = np.log2(col.bzero)
                if bits.is_integer():
                    dtype = BITS_TO_UINT[int(bits+1)]

            if hasattr(dtype, 'subdtype'):
                dtype = dtype.base

            qtable[col.name] = qtable[col.name].astype(dtype)

    return qtable


class ProductFactory(BasicRegistrationFactory):
    """
    A factory for generating stix data products

    Parameters
    ----------
    \\*inputs
        Inputs to parse for map objects. See the examples section for a
        detailed list of accepted inputs.

    Returns
    -------
    `stixpy.product.Product`
        If the input results in a singular product, then that is returned.
    `list` of `~stixpy.products.Product`
        If multiple inputs a list of `~stixpy.products.Product` objects will be
        returned.

    Examples
    --------

    """

    def _read_file(self, fname, **kwargs):
        """
        Read in a file name and return the list of (data, meta) pairs in that file.
        """
        # File gets read here.
        # NOTE: use os.fspath so that fname can be either a str or pathlib.Path
        # This can be removed once read_file supports pathlib.Path
        logger.debug(f'Reading {fname}')
        name = str(fname)
        if not name.endswith('.fits'):
            raise ValueError(f'File {fname} is not a fits file, only fits are supported')
        try:
            hdul = fits.open(fname, **kwargs)
        except Exception as e:
            msg = f"Failed to read {fname}."
            raise IOError(msg) from e

        if hdul[0].header.get('INSTRUME', '') != 'STIX':
            raise FileError(f"File '{fname}' is not a STIX fits file.")

        data = {'meta': hdul[0].header}
        for name in ['CONTROL', 'DATA', 'IDB_VERSIONS', 'ENERGIES']:
            try:
                data[name.lower()] = read_qtable(fname, hdu=name)
            except KeyError as e:
                if name in ('IDB_VERSIONS', 'ENERGIES'):
                    logger.debug(f"Extension '{name}' not in file '{fname}'")
                else:
                    raise e

        return [data]

    def _validate_meta(self, meta):
        """
        Validate a meta argument.
        """
        if isinstance(meta, astropy.io.fits.header.Header):
            return True
        elif isinstance(meta, dict):
            return True
        else:
            return False

    def _parse_args(self, *args, silence_errors=False, **kwargs):
        """
        Parses an args list into data-header pairs.
        args can contain any mixture of the following entries:
        * filename, as a str or pathlib.Path, which will be read
        * directory, as a str or pathlib.Path, from which all files will be read
        * glob, from which all files will be read
        * url, which will be downloaded and read
        * lists containing any of the above.
        Examples
        --------
        self._parse_args(['file1', 'file2', 'file3'],
                         'file4',
                         'directory1',
                         '*.fits')
        """
        # Account for nested lists of items
        args = expand_list(args)

        # Sanitise the input so that each 'type' of input corresponds to a different
        # class, so single dispatch can be used later
        nargs = len(args)
        i = 0
        while i < nargs:
            arg = args[i]
            if isinstance(arg, str) and is_url(arg):
                # Repalce URL string with a Request object to dispatch on later
                args[i] = Request(arg)
            elif possibly_a_path(arg):
                # Repalce path strings with Path objects
                args[i] = pathlib.Path(arg)
            i += 1

        # Parse the arguments
        # Note that this list can also contain GenericProducts if they are directly given to the factory
        meta_data_dicts = []
        for arg in args:
            try:
                meta_data_dicts += self._parse_arg(arg, **kwargs)
            except NoMapsInFileError as e:
                if not silence_errors:
                    raise
                warn_user(f"One of the arguments failed to parse with error: {e}")

        return meta_data_dicts

    # Note that post python 3.8 this can be @functools.singledispatchmethod
    @seconddispatch
    def _parse_arg(self, arg, **kwargs):
        """
        Take a factory input and parse into (data, header) pairs.
        Must return a list, even if only one pair is returned.
        """
        raise ValueError(f"Invalid input: {arg}")

    @_parse_arg.register(tuple)
    def _parse_tuple(self, arg, **kwargs):
        meta, control, data, *other = arg

        if self._validate_meta(meta):
            pair = (meta, control, data, *other)
        return [pair]

    # @_parse_arg.register(DatabaseEntryType)
    # def _parse_dbase(self, arg, **kwargs):
    #     return self._read_file(arg.path, **kwargs)

    @_parse_arg.register(GenericProduct)
    def _parse_map(self, arg, **kwargs):
        return [arg]

    @_parse_arg.register(Request)
    def _parse_url(self, arg, **kwargs):
        url = arg.full_url
        path = str(cache.download(url).absolute())
        pairs = self._read_file(path, **kwargs)
        return pairs

    @_parse_arg.register(pathlib.Path)
    def _parse_path(self, arg, **kwargs):
        return parse_path(arg, self._read_file, **kwargs)

    def __call__(self, *args, composite=False, sequence=False, silence_errors=False, **kwargs):
        """ Method for running the factory. Takes arbitrary arguments and
        keyword arguments and passes them to a sequence of pre-registered types
        to determine which is the correct Map-type to build.
        Arguments args and kwargs are passed through to the validation
        function and to the constructor for the final type. For Map types,
        validation function must take a data-header pair as an argument.
        Parameters
        ----------
        composite : `bool`, optional
            Indicates if collection of maps should be returned as a `~sunpy.map.CompositeMap`.
            Default is `False`.
        sequence : `bool`, optional
            Indicates if collection of maps should be returned as a `sunpy.map.MapSequence`.
            Default is `False`.
        silence_errors : `bool`, optional
            If set, ignore data-header pairs which cause an exception.
            Default is ``False``.
        Notes
        -----
        Extra keyword arguments are passed through to `sunpy.io.read_file` such
        as `memmap` for FITS files.
        """
        meta_control_data_dicts = self._parse_args(*args, silence_errors=silence_errors, **kwargs)
        new_products = list()

        # Loop over each registered type and check to see if WidgetType
        # matches the arguments.  If it does, use that type.
        for meta_control_data in meta_control_data_dicts:
            if isinstance(meta_control_data, GenericProduct):
                new_products.append(meta_control_data)
                continue

            try:
                new_product = self._check_registered_widgets(**meta_control_data, **kwargs)
                new_products.append(new_product)
            except (NoMatchError, MultipleMatchError,
                    ValidationFunctionError) as e:
                if not silence_errors:
                    raise
                warn_user(f"One of the inputs failed to validate with: {e}")

        if not len(new_products):
            raise RuntimeError('No maps loaded')

        if len(new_products) == 1:
            return new_products[0]

        return new_products

    def _check_registered_widgets(self, *, meta, control, data, **kwargs):

        candidate_widget_types = list()

        for key in self.registry:

            # Call the registered validation function for each registered class
            if self.registry[key](meta=meta, **kwargs):
                candidate_widget_types.append(key)

        n_matches = len(candidate_widget_types)

        if n_matches == 0:
            if self.default_widget_type is None:
                raise NoMatchError("No types match specified arguments and no default is set.")
            else:
                candidate_widget_types = [self.default_widget_type]
        elif n_matches > 1:
            raise MultipleMatchError("Too many candidate types identified "
                                     f"({candidate_widget_types}). "
                                     "Specify enough keywords to guarantee unique type "
                                     "identification.")

        # Only one is found
        WidgetType = candidate_widget_types[0]

        return WidgetType(meta=meta, control=control, data=data, **kwargs)


class InvalidMapInput(ValueError):
    """Exception to raise when input variable is not a Map instance and does
    not point to a valid Map input file."""


class InvalidMapType(ValueError):
    """Exception to raise when an invalid type of map is requested with Map
    """


class FileError(ValueError):
    """Exception to raise when input is not valid STIX file
    """


Product = ProductFactory(registry=GenericProduct._registry, default_widget_type=GenericProduct,
                 additional_validation_functions=['is_datasource_for'])



from pathlib import Path

import pytest

from stixpy.data import test
from stixpy.product.sources import QLLightCurve, Spectrogram
from stixpy.product.product_factory import Product

from stixpy.science import *

def test_product_factory_file():
    prod = Product(test.STIX_SCI_XRAY_SPEC)
    assert isinstance(prod, Spectrogram)

@pytest.mark.remote_data
def test_product_factory_url():
    ql_lc = Product('http://pub099.cs.technik.fhnw.ch/data/fits/L1/2022/11/15/QL/solo_L1_stix-ql-lightcurve_20221115_V01.fits')
    assert isinstance(ql_lc, QLLightCurve)

@pytest.mark.skip
@pytest.mark.remote_data
def test_product_factory_directory():
    test_data_path = 'http://pub099.cs.technik.fhnw.ch/data/fits/L1/2022/11/15/QL/' #Path(__file__).parent.parent.parent / 'data'
    products = Product(test_data_path, silence_errors=True)
    assert products




import pytest
import astropy.units as u
import numpy as np

from astropy.time import Time
from numpy.testing import assert_equal

from stixpy.utils.time import times_to_indices

@pytest.mark.parametrize('duration,expected_indices', [(5*u.s,  [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
                                                       (10*u.s, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
                                                       (15*u.s, [0, 1, 3, 4, 6, 7, 9]),
                                                       (20*u.s, [0, 2, 4, 6, 8, 9]),
                                                       (200*u.s, [0,9])])
def test_times_to_indices_scalar(duration, expected_indices):
    observed_times = Time('2021-12-20T00:00:00.0') + np.arange(10) * 10*u.s
    indices = times_to_indices(duration, observed_times)
    assert_equal(indices, np.array(expected_indices))

def test_times_to_indices_1d_times():
    observed_times = Time.now() + np.arange(10) * 10 * u.s
    request_times = observed_times[:]
    indices = times_to_indices(request_times, observed_times)
    assert_equal(indices, np.arange(10))
    indices1 = times_to_indices(request_times - 5.0 * u.s, observed_times)
    assert_equal(indices1, np.arange(9))
    indices2 = times_to_indices(request_times + 5 * u.s, observed_times)
    assert_equal(indices2, np.arange(10))

    uneven_obs_time = observed_times[0] + [5,  15,  30,  50,  70,  85,  95, 100] * u.s
    request_times = observed_times[0] + np.cumsum(np.arange(8)*10) * u.s
    indices3 = times_to_indices(request_times, uneven_obs_time)
    assert_equal(indices3, [0, 2, 3, 7])

def test_times_to_indices_2d_times():
    observed_times = Time.now() + np.arange(10) * 10 * u.s
    tmp = np.arange(10)*10
    requested_times = (observed_times[0]+[0, 90]*u.s).reshape(2, 1)
    indices = times_to_indices(requested_times, observed_times)
    assert_equal(indices, np.array([[0, 9]]))
    requested_times = observed_times[0] + np.vstack((tmp[0:-1], tmp[1:]))*u.s

    indices = times_to_indices(requested_times, observed_times)
    assert_equal(indices,  np.array([[0, 1], [1, 2], [2, 3], [3, 4], [4, 5],
                                  [5, 6], [6, 7], [7, 8], [8, 9]]))
    requested_times = observed_times[0] + [[0, 18], [18, 27], [27, 30], [30, 50], [50, 62],[62, 110]]*u.s
    indices2 = times_to_indices(requested_times, observed_times)
    assert_equal(indices2, np.array([[0, 2], [2, 3], [3, 3], [3, 5], [5, 6], [6, 9]]))
